(* ******************************************************************** *)
(*
 * Project         : UPF - A Uniform Policy Framework  Isabelle/HOL
 * Version         : 1.0
 *
 * Author          : Burkhart Wolff
 *
 * This file       : Core Theory
 *
 * Copyright (c) 2010 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: ROOT.ML 8591 2009-09-26 09:58:14Z brucker $ *)

header{* The Unified Policy Framework (UPF)*}

theory  UPFCore
imports Main
begin


section{* Foundation *}

text{* The purpose of this theory is to formalize a somewhat non-standard view
on the fundamental concept of a security policy which is worth outlining.
This view has arisen from prior experience in the modelling of network
(firewall) policies. Instead of regarding policies as relations on resources,
sets of permissions, etc., we emphasise the view that a policy is a policy
decision function that grants or denies access to resources, permissions, etc.
In other words, we model the concrete function that implements the policy
decision point in a system. An advantage of this view is that it is compatible
with many different policy models, enabling a uniform modelling framework to be
defined. Furthermore, this function is typically a large cascade of nested
conditionals, using conditions referring to an internal state and security
contexts of the system or a user. This cascade of conditionals can easily be
decomposed into a set of test cases similar to transformations used for binary
decision diagrams (BDD). From the modelling perspective, using HOL as
its input language, we will consequently use the expressive power of its
underlying functional  programming language, including the possibility to
define higher-order combinators.

In more detail, we model policies as partial functions based on input
data $\alpha$ (arguments, system state, security context, ...) to output
data $\beta$: *}
datatype '\<alpha> decision = allow '\<alpha>  | deny '\<alpha>

type_synonym ('\<alpha>,'\<beta>) policy = "'\<alpha>  \<rightharpoonup> '\<beta> decision" (infixr "|->" 0)
translations (type) "'\<alpha> |-> '\<beta>" <= (type) "'\<alpha>  \<rightharpoonup> '\<beta> decision"


type_notation (xsymbols)
  "policy" (infixr "\<mapsto>" 0) (*TODO: Better Notation *)

text{*Thus, the range of a policy may consist of \verb+Some(accept x)+ data,
of \verb+Some(deny x)+ data, as well as \verb+None+ modeling the undefinedness
of a policy, \ie() a policy is considered as a partial function. Partial 
functions are used since we describe elementary policies by partial system 
behaviour, which are glued together by operators such as function override and 
functional composition. *}

text{* We define the two fundamental sets, the allow-set $Allow$ and the
deny-set $Deny$ (written $A$ and $D$ set for short), to characterize these
two main sets of the range of a policy. *}

definition Allow :: "('\<alpha> decision) set"
where     "Allow \<equiv> range allow"

definition Deny :: "('\<alpha> decision) set"
where     "Deny \<equiv> range deny"

text{* A key concept is the flattening of nested decisions.
  There are 4 possibilities, and these possibilities will
  give the various flavours of policy composition. *}
fun    flat_orA :: "('\<alpha> decision) decision \<Rightarrow> ('\<alpha> decision)"
where "flat_orA(allow(allow y)) = allow y"
     |"flat_orA(allow(deny y))  = allow y"
     |"flat_orA(deny(allow y))  = allow y"
     |"flat_orA(deny(deny y))   = deny y"

lemma flat_orA_deny[dest]:"flat_orA x = deny y \<Longrightarrow> x = deny(deny y)"
apply(case_tac "x", simp_all)
apply(case_tac "\<alpha>", simp_all)
apply(case_tac "\<alpha>", simp_all)
done

lemma flat_orA_allow[dest]:
  "flat_orA x = allow y \<Longrightarrow> x = allow(allow y) 
                          \<or> x = allow(deny y) 
                          \<or> x = deny(allow y)"
apply(case_tac "x", simp_all)
apply(case_tac "\<alpha>", simp_all)
apply(case_tac "\<alpha>", simp_all)
done

fun    flat_orD :: "('\<alpha> decision) decision \<Rightarrow> ('\<alpha> decision)"
where "flat_orD(allow(allow y)) = allow y"
     |"flat_orD(allow(deny y))  = deny y"
     |"flat_orD(deny(allow y))  = deny y"
     |"flat_orD(deny(deny y))   = deny y"

lemma flat_orD_allow[dest]:"flat_orD x = allow y \<Longrightarrow> x = allow(allow y)"
apply(case_tac "x", simp_all)
apply(case_tac "\<alpha>", simp_all)
apply(case_tac "\<alpha>", simp_all)
done

lemma flat_orD_deny[dest]:
  "flat_orD x = deny y \<Longrightarrow>  x = deny(deny y) 
                          \<or> x = allow(deny y) 
                          \<or> x = deny(allow y)"
apply(case_tac "x", simp_all)
apply(case_tac "\<alpha>", simp_all)
apply(case_tac "\<alpha>", simp_all)
done

fun    flat_1 :: "('\<alpha> decision) decision \<Rightarrow> ('\<alpha> decision)"
where "flat_1(allow(allow y)) = allow y"
     |"flat_1(allow(deny y))  = allow y"
     |"flat_1(deny(allow y))  = deny y"
     |"flat_1(deny(deny y))   = deny y"

lemma flat_1_allow[dest]:
   "flat_1 x = allow y \<Longrightarrow> x = allow(allow y) \<or> x = allow(deny y)"
apply(case_tac "x", simp_all)
apply(case_tac "\<alpha>", simp_all)
apply(case_tac "\<alpha>", simp_all)
done

lemma flat_1_deny[dest]:
  "flat_1 x = deny y \<Longrightarrow>  x = deny(deny y) \<or> x = deny(allow y)"
apply(case_tac "x", simp_all)
apply(case_tac "\<alpha>", simp_all)
apply(case_tac "\<alpha>", simp_all)
done


fun    flat_2 :: "('\<alpha> decision) decision \<Rightarrow> ('\<alpha> decision)"
where "flat_2(allow(allow y)) = allow y"
     |"flat_2(allow(deny y))  = deny y"
     |"flat_2(deny(allow y))  = allow y"
     |"flat_2(deny(deny y))   = deny y"


lemma flat_2_allow[dest]:
   "flat_2 x = allow y \<Longrightarrow> x = allow(allow y) \<or> x = deny(allow y)"
apply(case_tac "x", simp_all)
apply(case_tac "\<alpha>", simp_all)
apply(case_tac "\<alpha>", simp_all)
done

lemma flat_2_deny[dest]:
  "flat_2 x = deny y \<Longrightarrow>  x = deny(deny y) \<or> x = allow(deny y)"
apply(case_tac "x", simp_all)
apply(case_tac "\<alpha>", simp_all)
apply(case_tac "\<alpha>", simp_all)
done
 

subsection{* Policy Constructors *}

text{* Most elementary policy constructors are based on the
  update operation \verb+Fun.fun_upd_def+ {@thm Fun.fun_upd_def}
  and the maplet-notation $a(x \maplet y)$ used for $a(x := Some y)$.

  However, we add notation on top of this adopted to our problem
  domain: 
*}

nonterminal policylets and policylet

syntax
  "_policylet1"  :: "['a, 'a] => policylet"             ("_ /+=/ _")
  "_policylet2"  :: "['a, 'a] => policylet"             ("_ /-=/ _")
  ""         :: "policylet => policylets"               ("_")
  "_Maplets" :: "[policylet, policylets] => policylets" ("_,/ _")
  "_Maplets" :: "[policylet, policylets] => policylets" ("_,/ _")
  "_MapUpd"  :: "['a |-> 'b, policylets] => 'a |-> 'b"  ("_/'(_')" [900,0]900)

syntax (xsymbols)
  "_policylet1"  :: "['a, 'a] => policylet"             ("_ /+\<mapsto>/ _")
  "_policylet2"  :: "['a, 'a] => policylet"             ("_ /-\<mapsto>/ _")
  "_emptypolicy" :: "'a |-> 'b"                         ("\<emptyset>")

translations
  "_MapUpd m (_Maplets xy ms)"    == "_MapUpd (_MapUpd m xy) ms"
  "_MapUpd m (_policylet1  x y)"  == "m(x := CONST Some (CONST allow y))"
  "_MapUpd m (_policylet2  x y)"  == "m(x := CONST Some (CONST deny y))"
  "\<emptyset>"                             == "CONST empty"

lemma test: "empty(x+=a, y-= b) = \<emptyset>(x +\<mapsto> a, y -\<mapsto> b)" by simp

text{* We inherit a fairly rich theory on policy updates from
  Map here. Some examples are: *}

lemma pol_upd_triv1: "t k = Some (allow x) \<Longrightarrow> t(k+\<mapsto>x) = t"
  by (rule ext) simp
lemma pol_upd_triv2: "t k = Some (deny x) \<Longrightarrow> t(k-\<mapsto>x) = t"
  by (rule ext) simp

lemma pol_upd_allow_nonempty: "t(k+\<mapsto>x) \<noteq> \<emptyset>" by simp
lemma pol_upd_deny_nonempty: "t(k-\<mapsto>x) \<noteq> \<emptyset>" by simp

lemma pol_upd_eqD1 : "m(a+\<mapsto>x) = n(a+\<mapsto>y) \<Longrightarrow> x = y"
  by(auto dest: map_upd_eqD1)

lemma pol_upd_eqD2 : "m(a-\<mapsto>x) = n(a-\<mapsto>y) \<Longrightarrow> x = y"
  by(auto dest: map_upd_eqD1)

lemma pol_upd_neq1 [simp]: "m(a+\<mapsto>x) \<noteq> n(a-\<mapsto>y)"
  by(auto dest: map_upd_eqD1)


text{* Another key-constructor of policies is the policy-override operator
  inherited from the map-theory. Since it is an usual construct to
  build complex policies mimicking the style to compose policy rules
  in a "left-to-right-first-fit" - manner, we redefine the override
  from the map-theory (which does the trick in reverse order).)*}

syntax 
  "_policyoverride" :: "['a \<mapsto> 'b, 'a \<mapsto> 'b] \<Rightarrow> 'a \<mapsto> 'b" (infixl "(+p/)" 100)

syntax (xsymbols)
  "_policyoverride"  :: "['a \<mapsto> 'b, 'a \<mapsto> 'b] \<Rightarrow> 'a \<mapsto> 'b" (infixl "\<Oplus>" 100)

find_theorems "_ ++ _"
translations
  "p \<Oplus> q" == "q ++ p" 


text{* Some elementary facts inherited from Map are:*}

lemma override_empty: "p \<Oplus> \<emptyset> = p" by(simp)

lemma empty_override: "\<emptyset> \<Oplus> p = p" by(simp)

lemma override_assoc: "p1 \<Oplus> (p2 \<Oplus> p3) = (p1 \<Oplus> p2) \<Oplus> p3" by simp



text{* The following operators are slightly more ad-hoc variants of override. 
For \verb+override_A+, in the domain of the second argument, however, 
an allow of the first arg wins over a deny.  
For \verb+override_D+, the situation is dual. 
*}

definition override_A ::"['\<alpha>\<mapsto>'\<beta>, '\<alpha>\<mapsto>'\<beta>] \<Rightarrow> '\<alpha>\<mapsto>'\<beta>" (infixl "++'_A" 100) 
where  "m2 ++_A m1 = 
          (\<lambda>x. case m1 x of 
                 Some(allow a) \<Rightarrow>  Some (allow a) 
               | Some(deny a)  \<Rightarrow> (case m2 x of Some (allow b) \<Rightarrow> Some (allow b) 
                                            | _ \<Rightarrow> Some (deny a))
               | None \<Rightarrow> m2 x 
           )"

syntax (xsymbols)
  "_policyoverride_A"  :: "['a \<mapsto> 'b, 'a \<mapsto> 'b] \<Rightarrow> 'a \<mapsto> 'b" (infixl "\<Oplus>\<sub>A" 100)
translations
  "p \<Oplus>\<sub>A q" == "p ++_A q"

lemma override_A_empty[simp]: "p \<Oplus>\<sub>A \<emptyset> = p" by(simp add:override_A_def)

lemma empty_override_A[simp]: "\<emptyset> \<Oplus>\<sub>A p = p" 
  apply(rule ext, simp add:override_A_def)
  apply(case_tac "p x", simp_all)
  apply(case_tac a, simp_all)
  done


lemma override_A_assoc: 
  "p1 \<Oplus>\<sub>A (p2 \<Oplus>\<sub>A p3) = (p1 \<Oplus>\<sub>A p2) \<Oplus>\<sub>A p3" 
by (rule ext, simp add: override_A_def split: decision.splits  option.splits)

definition override_D :: "['\<alpha>\<mapsto>'\<beta>, '\<alpha>\<mapsto>'\<beta>] \<Rightarrow> '\<alpha>\<mapsto>'\<beta>" (infixl "++'_D" 100) 
where "m1 ++_D m2 = 
          (\<lambda>x. case m2 x of 
                Some(deny a) \<Rightarrow> Some (deny a) 
              | Some(allow a) \<Rightarrow> (case m1 x of Some (deny b) \<Rightarrow> Some (deny b) 
                                  | _ \<Rightarrow> Some (allow a))
              | None \<Rightarrow> m1 x 
           )"
 
syntax (xsymbols)
  "_policyoverride_D"  :: "['a \<mapsto> 'b, 'a \<mapsto> 'b] \<Rightarrow> 'a \<mapsto> 'b" (infixl "\<Oplus>\<sub>D" 100)
translations
  "p \<Oplus>\<sub>D q" == "p ++_D q"

lemma override_D_empty[simp]: "p \<Oplus>\<sub>D \<emptyset> = p" by(simp add:override_D_def)

lemma empty_override_D[simp]: "\<emptyset> \<Oplus>\<sub>D p = p" 
  apply(rule ext, simp add:override_D_def)
  apply(case_tac "p x", simp_all)
  apply(case_tac a, simp_all)
  done

lemma override_D_assoc: 
  "p1 \<Oplus>\<sub>D (p2 \<Oplus>\<sub>D p3) = (p1 \<Oplus>\<sub>D p2) \<Oplus>\<sub>D p3"
by (rule ext, simp add: override_D_def split: decision.splits  option.splits)



section{* Injections and Bijections of Policies *}

definition
   deny_pfun    :: "('\<alpha> \<rightharpoonup>'\<beta>) \<Rightarrow> ('\<alpha> \<mapsto> '\<beta>)" (binder "AllD" 10)
   where 
  "deny_pfun pf \<equiv> (\<lambda> x. case pf x of
                          Some y \<Rightarrow> Some(deny(y))
                         |None \<Rightarrow> None)"

definition
   allow_pfun    :: "('\<alpha> \<rightharpoonup>'\<beta>) \<Rightarrow> ('\<alpha> \<mapsto> '\<beta>)" (binder "AllA" 10)
   where 
  "allow_pfun pf \<equiv> (\<lambda> x. case pf x of
                          Some y \<Rightarrow> Some(allow(y))
                         |None \<Rightarrow> None)"

notation (xsymbols)
   deny_pfun  (binder "\<forall>D" 10) and
   allow_pfun (binder "\<forall>A" 10)
(* TODO: The LaTeX variant should be more like $\forall_D~$ 
   in order to increase readability \<dots> *)

lemma AllD_norm[simp]: "deny_pfun (id o Some) = (\<forall>Dx. Some x)"
by(simp add:id_def comp_def)

lemma AllD_norm2[simp]: "deny_pfun (Some o id) = (\<forall>Dx. Some x)"
by(simp add:id_def comp_def)

lemma AllA_norm[simp]: "allow_pfun (id o Some) = (\<forall>Ax. Some x)"
by(simp add:id_def comp_def)

lemma AllA_norm2[simp]: "allow_pfun (Some o id) = (\<forall>Ax. Some x)"
by(simp add:id_def comp_def)

lemma AllA_apply[simp]: "(\<forall>Ax. Some (P x)) x = Some(allow(P x))"
by(simp add:allow_pfun_def)

lemma AllD_apply[simp]: "(\<forall>Dx. Some (P x)) x = Some(deny(P x))"
by(simp add:deny_pfun_def)

text{* TODO : The syntax does not work yet. Actually,
I'd like to view these operators as quantifiers: 
$\forall_A x. E$ etc . . .*}

(*
lemma neq_Allow_Deny: "pf \<noteq> \<emptyset> \<Longrightarrow> (\<forall>D pf) \<noteq> (\<forall>N pf)"
*)
lemma neq_Allow_Deny: "pf \<noteq> \<emptyset> \<Longrightarrow> (deny_pfun pf) \<noteq> (allow_pfun pf)"
apply(erule contrapos_nn, rule ext)
apply(drule_tac x=x in fun_cong)
apply(auto simp: deny_pfun_def allow_pfun_def)
apply(case_tac "pf x =  None", auto)
done

text{* Some common instances : *}


definition allow_all_general :: "('\<alpha> \<Rightarrow> '\<beta>) \<Rightarrow> ('\<alpha> \<mapsto> '\<beta>)" 
  where "allow_all_general f \<equiv> allow_pfun (Some o f)"

definition deny_all_general :: "('\<alpha> \<Rightarrow> '\<beta>) \<Rightarrow> ('\<alpha> \<mapsto> '\<beta>)" 
  where "deny_all_general f \<equiv> deny_pfun (Some o f)"


definition
   deny_all_id    :: "'\<alpha> \<mapsto> '\<alpha>" where
  "deny_all_id  \<equiv> deny_pfun (id o Some)"

definition
   allow_all_id    :: "'\<alpha> \<mapsto> '\<alpha>" where
  "allow_all_id  \<equiv> allow_pfun (id o Some)"

text{* ... and resulting properties: *}

lemma "allow_all_id  \<Oplus> empty  = allow_all_id"
apply simp
done

lemma "allow_pfun empty = empty"
apply (rule ext)
apply (simp add: allow_pfun_def)
done


lemma allow_left_cancel :"dom pf = UNIV \<Longrightarrow> (allow_pfun pf) \<Oplus> x = (allow_pfun pf)" 
apply (rule ext)+
apply (auto simp: allow_pfun_def option.splits)
done


lemma deny_left_cancel :"dom pf = UNIV \<Longrightarrow> (deny_pfun pf) \<Oplus> x = (deny_pfun pf)"
apply (rule ext)+
apply (auto simp: deny_pfun_def option.splits)
done

text{* In the following, we establish an isomorpism between 
  transition policies and a state exception monads. Recall: *}
type_synonym ('o, '\<sigma>) MON_SE = "'\<sigma> \<rightharpoonup> ('o \<times> '\<sigma>)" 

definition policy2MON ::"('\<iota>\<times>'\<sigma> \<mapsto> 'o\<times>'\<sigma>) \<Rightarrow> ('\<iota> \<Rightarrow>('o decision,'\<sigma>)MON_SE)"
where "policy2MON p = (\<lambda> \<iota> \<sigma>. case p (\<iota>,\<sigma>) of
                              Some (allow (out,\<sigma>')) \<Rightarrow> (Some(allow out, \<sigma>'))
                            | Some (deny (out,\<sigma>')) \<Rightarrow> (Some(deny out, \<sigma>'))
                            | None \<Rightarrow> None)"
(* WOW ! This nested pattern stuff works !*)


section{* Domain, Range, and Restictions *}

find_theorems name:"range"
text{*
  Since policies are essentially maps, we inherit the basic definitions
  for domain and range on  Maps: \\
  \verb+Map.dom_def+ :  {@thm Map.dom_def} \\
  whereas range is just an abrreviation for image:
  \begin{verbatim}
  abbreviation range :: "('a => 'b) => 'b set" 
  where -- "of function"  "range f == f ` UNIV"
  \end{verbatim}
  As a consequence, we inherit the following properties on
  policies:
  \begin{itemize}
  \item  \verb+Map.domD+ {@thm Map.domD}
  \item\verb+Map.domI+ {@thm Map.domI}
  \item\verb+Map.domIff+ {@thm Map.domIff}
  \item\verb+Map.dom_const+ {@thm Map.dom_const}
  \item\verb+Map.dom_def_raw+ {@thm Map.dom_def_raw}
  \item\verb+Map.dom_empty+ {@thm Map.dom_empty}
  \item\verb+Map.dom_eq_empty_conv+ {@thm Map.dom_eq_empty_conv}
  \item\verb+Map.dom_eq_singleton_conv+ {@thm Map.dom_eq_singleton_conv}
  \item\verb+Map.dom_fun_upd+ {@thm Map.dom_fun_upd}
  \item\verb+Map.dom_if+ {@thm Map.dom_if}
  \item\verb+Map.dom_map_add+ {@thm Map.dom_map_add}
  \end{itemize}
*}

text{* However, some properties are specific to policy concepts: *}
lemma sub_ran : "ran p  \<subseteq> Allow \<union> Deny"
apply(auto simp: Allow_def Deny_def ran_def full_SetCompr_eq[symmetric])
apply(case_tac "x", simp_all)
apply(erule_tac x="\<alpha>" in allE, simp)
done

lemma dom_allow_pfun [simp]:"dom(allow_pfun f) = dom f"
apply(auto simp: allow_pfun_def)
apply(case_tac "f x", simp_all)
done

lemma dom_allow_all: "dom(allow_all_general f) = UNIV"
by(auto simp: allow_all_general_def o_def)

lemma dom_deny_pfun [simp]:"dom(deny_pfun f) = dom f"
apply(auto simp: deny_pfun_def)
apply(case_tac "f x", simp_all)
done

lemma dom_deny_all: " dom(deny_all_general f) = UNIV"
by(auto simp: deny_all_general_def o_def)


lemma ran_allow_pfun [simp]:"ran(allow_pfun f) = allow `(ran f)"
apply(simp add: allow_pfun_def ran_def) 
apply(rule set_eqI, auto)
apply(case_tac "f a", auto simp: image_def)
apply(rule_tac x=a in exI, simp)
done

lemma ran_allow_all: "ran(allow_all_general id) = Allow"
apply(simp add: allow_all_general_def Allow_def o_def)
apply(rule set_eqI, auto simp: image_def ran_def)
done

lemma ran_deny_pfun[simp]: "ran(deny_pfun f) = deny ` (ran f)"
apply(simp add: deny_pfun_def ran_def)
apply(rule set_eqI, auto)
apply(case_tac "f a", auto simp: image_def)
apply(rule_tac x=a in exI, simp)
done

lemma ran_deny_all: "ran(deny_all_general id) = Deny"
apply(simp add: deny_all_general_def Deny_def o_def)
apply(rule set_eqI, auto simp: image_def ran_def)
done

text{* Reasoning over \verb+dom+ is most crucial since it paves
the way for simplification and reordering of policies composed
by override (\ie{}, by the normal left-to-right rule composition
method.

\begin{itemize}
\item \verb+Map.dom_map_add+ {@thm Map.dom_map_add}
\item \verb+Map.inj_on_map_add_dom+ {@thm +Map.inj_on_map_add_dom}
\item \verb+Map.map_add_comm+ {@thm Map.map_add_comm}
\item \verb+Map.map_add_dom_app_simps(1)+ {@thm Map.map_add_dom_app_simps(1)}
\item \verb+Map.map_add_dom_app_simps(2)+ {@thm Map.map_add_dom_app_simps(2)}
\item \verb+Map.map_add_dom_app_simps(3)+ {@thm Map.map_add_dom_app_simps(3)}
\item \verb+Map.map_add_upd_left+ {@thm Map.map_add_upd_left}
\end{itemize}
The latter rule also applies to allow- and deny-override.
*}


definition dom_restrict :: "['\<alpha> set, '\<alpha>\<mapsto>'\<beta>] \<Rightarrow> '\<alpha>\<mapsto>'\<beta>" (infixr "\<triangleleft>" 55)
where     "S \<triangleleft> p \<equiv> (\<lambda>x. if x \<in> S then p x else None)"

lemma dom_dom_restrict[simp] : "dom(S \<triangleleft> p) = S \<inter> dom p"
apply(auto simp: dom_restrict_def)
apply(case_tac "x \<in> S", simp_all)
apply(case_tac "x \<in> S", simp_all)
done

lemma dom_restrict_idem[simp] : "(dom p) \<triangleleft> p = p"
by(rule ext, 
   auto simp: dom_restrict_def
        dest: neq_commute[THEN iffD1,THEN not_None_eq [THEN iffD1]])

lemma dom_restrict_inter[simp] : "T \<triangleleft> S \<triangleleft> p = T \<inter> S \<triangleleft> p"
by(rule ext, 
   auto simp: dom_restrict_def
        dest: neq_commute[THEN iffD1,THEN not_None_eq [THEN iffD1]])


definition ran_restrict :: "['\<alpha>\<mapsto>'\<beta>,'\<beta> decision set] \<Rightarrow> '\<alpha>\<mapsto>'\<beta>" (infixr "\<triangleright>" 55)
where     "p \<triangleright> S \<equiv> (\<lambda>x. if p x \<in> (Some`S) then p x else None)"

lemma ran_ran_restrict[simp] : "ran(p \<triangleright> S) = S \<inter> ran p"
by(auto simp: ran_restrict_def image_def ran_def)

lemma ran_restrict_idem[simp] : "p \<triangleright> (ran p) = p"
apply(rule ext, auto simp: ran_restrict_def image_def Ball_def ran_def)
apply(erule contrapos_pp)
apply(auto dest!: neq_commute[THEN iffD1,THEN not_None_eq [THEN iffD1]])
done

lemma ran_restrict_inter[simp] : "(p \<triangleright> S) \<triangleright> T = p \<triangleright> T \<inter> S"
by(rule ext, 
   auto simp: ran_restrict_def
        dest: neq_commute[THEN iffD1,THEN not_None_eq [THEN iffD1]])

lemma ran_gen_A[simp] : "(\<forall>Ax. Some(P x)) \<triangleright> Allow = (\<forall>Ax. Some(P x))"
by(rule ext, auto simp: Allow_def ran_restrict_def)
 
lemma ran_gen_D[simp] : "(\<forall>Dx. Some(P x)) \<triangleright> Deny = (\<forall>Dx. Some(P x))"
by(rule ext, auto simp: Deny_def ran_restrict_def)

section{* Fundamental Compositions *}

text{* Of course, the domain of a policy can be adapted via a conventional
  function composition: *}

lemma test :  "(p::'\<beta> \<mapsto> '\<gamma>) o (f::'\<alpha> \<Rightarrow> '\<beta>) = (C::'\<alpha> \<mapsto> '\<gamma>)"
oops

text{* An slightly more complicated analogon for the range of a policy
  is defined as follows: *}

definition policy_range_comp :: "['\<beta>\<Rightarrow>'\<gamma>, '\<alpha>\<mapsto>'\<beta>] \<Rightarrow> '\<alpha> \<mapsto>'\<gamma>"  
                                                    (infixl "o'_f" 55) 
where
  "f o_f p \<equiv> \<lambda>x. case p x of
                     Some(allow y) \<Rightarrow> Some(allow (f y))
                   | Some(deny y) \<Rightarrow> Some(deny (f y))
                   | None \<Rightarrow> None"

lemma policy_range_comp_strict : "f o_f \<emptyset> = \<emptyset>"
by(rule ext, simp add: policy_range_comp_def)

definition policy_range_comp2 :: "['\<beta>\<rightharpoonup>'\<gamma>,'\<alpha>\<mapsto>'\<beta>]\<Rightarrow>'\<alpha>\<mapsto>'\<gamma>" (infixl "o'_pf" 55) 
where
  "f o_pf p \<equiv> \<lambda>x. case p x of
                     Some(allow y) \<Rightarrow> (case f y of
                                        Some y' \<Rightarrow> Some(allow y')
                                      | None \<Rightarrow> None)
                   | Some(deny y) \<Rightarrow> (case f y of 
                                       Some y' \<Rightarrow> Some(deny y')
                                      | None \<Rightarrow> None)
                   | None \<Rightarrow> None"

lemma policy_range_comp2_strict : "pf o_pf \<emptyset> = \<emptyset>"
by(rule ext, simp add: policy_range_comp2_def)


lemma policy_range_comp2_strict2 : "\<emptyset> o_pf p = \<emptyset>"
apply(rule ext, simp add: policy_range_comp2_def)
apply(case_tac "p x", simp_all)
apply(case_tac "a", simp_all)
done

text{* A generalized version is: *}

definition range_split :: "[('\<beta>\<Rightarrow>'\<gamma>)\<times>('\<beta>\<Rightarrow>'\<gamma>),'\<alpha> \<mapsto> '\<beta>] \<Rightarrow> '\<alpha> \<mapsto> '\<gamma>"
                          (infixr "\<nabla>" 100)
where "(P) \<nabla> p \<equiv> \<lambda>x. case p x of 
                          Some(allow y) \<Rightarrow> Some(allow ((fst P) y))
                        | Some(deny y) \<Rightarrow> Some(deny ((snd P) y)) 
                        | None \<Rightarrow> None"

lemma range_split_strict[simp]: "P \<nabla> \<emptyset> = \<emptyset>"
by(rule ext, simp add: range_split_def)

lemma range_split_charn:
        "(f,g) \<nabla> p = (\<lambda>x. case p x of 
                           Some(allow x) \<Rightarrow> Some(allow (f x))
                         | Some(deny x) \<Rightarrow> Some(deny (g x)) 
                         | None \<Rightarrow> None)"
apply(simp add: range_split_def, rule ext)
apply(case_tac "p x", simp_all)
apply(case_tac "a", simp_all)
done

text{* The connection between these two becomes apparent if
  considering the following lemma:*}

lemma range_split_vs_range_compose: "(f,f) \<nabla> p = f o_f p"
by(simp add: range_split_charn policy_range_comp_def)

lemma range_split_id [simp] : "(id,id) \<nabla> p = p"
apply(rule ext, simp add: range_split_charn id_def)
apply(case_tac "p x", simp_all)
apply(case_tac "a", simp_all)
done

lemma range_split_bi_compose [simp]: 
  "(f1,f2) \<nabla> (g1,g2) \<nabla> p = (f1 o g1,f2 o g2) \<nabla> p"
apply(rule ext, simp add: range_split_charn comp_def)
apply(case_tac "p x", simp_all)
apply(case_tac "a", simp_all)
done


text {* The following is a variant of range_split, where the change in
the decision depends on the input instead of the output. *}


definition dom_split2 :: "[('\<alpha> \<Rightarrow> '\<gamma>) \<times> ('\<alpha> \<Rightarrow>'\<gamma>),'\<alpha> \<mapsto> '\<beta>] \<Rightarrow> '\<alpha> \<mapsto> '\<gamma>"
                          (infixr "\<Delta>" 100)
where "(P) \<Delta> p \<equiv> \<lambda>x. case p x of 
                          Some(allow y) \<Rightarrow> Some(allow ((fst P) x))
                        | Some(deny y) \<Rightarrow> Some(deny ((snd P) x)) 
                        | None \<Rightarrow> None"

lemma dom_split_alt_def :
  "(f, g) \<Delta> p = (dom(p \<triangleright> Allow) \<triangleleft> (allow_all_general f)) \<Oplus>
                 (dom(p \<triangleright> Deny) \<triangleleft> (deny_all_general g))"
apply(rule ext)
apply(simp add: dom_split2_def Allow_def Deny_def dom_restrict_def
                deny_all_general_def allow_all_general_def map_add_def)
apply(simp split:option.splits decision.splits)
apply(auto simp:map_add_def o_def deny_pfun_def ran_restrict_def image_def)
done



definition range_split2 :: "[('\<alpha> \<Rightarrow> '\<gamma>) \<times> ('\<alpha> \<Rightarrow>'\<gamma>),'\<alpha> \<mapsto> '\<beta>] \<Rightarrow> '\<alpha> \<mapsto> ('\<beta> \<times>'\<gamma>)"
                          (infixr "\<nabla>2" 100)
where "(P) \<nabla>2 p \<equiv> \<lambda>x. case p x of 
                          Some(allow y) \<Rightarrow> Some(allow (y,(fst P) x))
                        | Some(deny y) \<Rightarrow> Some(deny (y,(snd P) x)) 
                        | None \<Rightarrow> None"



text{* A common case: only for allow results were propagated: *}

lemma "(Some,K None) \<nabla> p = (f::'a \<mapsto> 'b option)"
oops

text{* Another common case considers transition policies: 
       only allow results were propagated, however, in 
       both cases, the state is propagated. *}

lemma "(\<lambda>(\<iota>,\<sigma>).(Some \<iota>,\<sigma>),\<lambda>(\<iota>,\<sigma>).(None, \<sigma>)) \<nabla> p = (f::'\<alpha>\<times>'\<sigma> \<mapsto> '\<beta> option \<times>'\<sigma>)"
oops


(* AB HIER BEGINNT DIE BAUSTELLE !!! WAS SOLL, KONZEPTIONELL GESEHEN,
   DAS H ??? ICH VERMUTE, ES FEHLT NEBEN DEN INJEKTOREN
   $\forall D$ etc. AUCH DIE INVERSEN !!! *)

text{* The following definition allows to compose two policies.
Denies and allows are transferred. *}

fun lift :: "('\<alpha> \<mapsto> '\<beta>) \<Rightarrow> ('\<alpha> decision \<mapsto>'\<beta> decision)"  
where "lift f (deny s)  = (case f s of 
                             Some y \<Rightarrow> Some(deny(y))
                           | None \<Rightarrow> None)"
    | "lift f (allow s) = (case f s of 
                             Some y \<Rightarrow> Some(allow y)
                           | None \<Rightarrow> None)"

lemma lift_mt [simp]: "lift \<emptyset> = \<emptyset>"
by(rule ext, case_tac "x",simp_all)  

(*
lemma lift_abstract1 : 
  "lift f (deny s)  = H (Option.map (deny o f))"
oops
(* there should be ways to express this in more abstract ways . . .*) 
*)

text{* Since policies are maps, we inherit a composition on
  them. However, this results in nestings of decisions --- 
  which must be flattened. As we now that there are 4
  different forms of flattening, we have four different
  forms of policy composition: *}

definition
  comp_orA :: "['\<beta>\<mapsto>'\<gamma>, '\<alpha>\<mapsto>'\<beta>] \<Rightarrow> '\<alpha>\<mapsto>'\<gamma>"  (infixl "o'_orA" 55) where
  "p2 o_orA p1 \<equiv> (Option.map flat_orA) o (lift (p2) o_m (p1))"


notation (xsymbols)
  comp_orA  (infixl "\<circ>\<^sub>\<or>\<^sub>A" 55)


lemma comp_orA_mt[simp]:"p \<circ>\<^sub>\<or>\<^sub>A \<emptyset> = \<emptyset>"
by(simp add: comp_orA_def)

lemma mt_comp_orA[simp]:"\<emptyset> \<circ>\<^sub>\<or>\<^sub>A p = \<emptyset>"
by(simp add: comp_orA_def)

lemma comp_orA_assoc:
  "p3 o_orA (p2 o_orA p1) = 
   p3 o_orA p2 o_orA p1"
oops



definition
  comp_orD :: "['\<beta>\<mapsto>'\<gamma>, '\<alpha>\<mapsto>'\<beta>] \<Rightarrow> '\<alpha>\<mapsto>'\<gamma>"  (infixl "o'_orD" 55) where
  "p2 o_orD p1 \<equiv> (Option.map flat_orD) o (lift (p2) o_m (p1))"

notation (xsymbols)
  comp_orD  (infixl "\<circ>\<^sub>orD" 55)

lemma comp_orD_mt[simp]:"p o_orD \<emptyset> = \<emptyset>"
by(simp add: comp_orD_def)

lemma mt_comp_orD[simp]:"\<emptyset> o_orD p = \<emptyset>"
by(simp add: comp_orD_def)


lemma comp_orD_assoc:
  "p3 o_orD (p2 o_orD p1) = p3 o_orD p2 o_orD p1"
apply(simp add: comp_orD_def)
oops


definition
  comp_1 :: "['\<beta>\<mapsto>'\<gamma>, '\<alpha>\<mapsto>'\<beta>] \<Rightarrow> '\<alpha>\<mapsto>'\<gamma>"  (infixl "o'_1" 55) where
  "p2 o_1 p1 \<equiv> (Option.map flat_1) o (lift (p2) o_m (p1))"

notation (xsymbols)
  comp_1  (infixl "\<circ>\<^sub>1" 55)

lemma comp_1_mt[simp]:"p \<circ>\<^sub>1 \<emptyset> = \<emptyset>"
by(simp add: comp_1_def)

lemma mt_comp_1[simp]:"\<emptyset> \<circ>\<^sub>1 p = \<emptyset>"
by(simp add: comp_1_def)

lemma comp_1_assoc:
  "p3 \<circ>\<^sub>1 (p2 \<circ>\<^sub>1 p1) = p3 \<circ>\<^sub>1 p2 \<circ>\<^sub>1 p1"
apply(simp add: comp_1_def)
oops

definition
  comp_2 :: "['\<beta>\<mapsto>'\<gamma>, '\<alpha>\<mapsto>'\<beta>] \<Rightarrow> '\<alpha>\<mapsto>'\<gamma>"  (infixl "o'_2" 55) where
  "p2 o_2 p1 \<equiv> (Option.map flat_2) o (lift (p2) o_m (p1))"

notation (xsymbols)
  comp_2  (infixl "\<circ>\<^sub>2" 55)

lemma comp_2_mt[simp]:"p \<circ>\<^sub>2 \<emptyset> = \<emptyset>"
by(simp add: comp_2_def)

lemma mt_comp_2[simp]:"\<emptyset> \<circ>\<^sub>2 p = \<emptyset>"
by(simp add: comp_2_def)

lemma comp_2_assoc:
  "p3 \<circ>\<^sub>2 (p2 \<circ>\<^sub>2 p1) = p3 \<circ>\<^sub>2 p2 \<circ>\<^sub>2 p1"
apply(simp add: comp_2_def)
oops


section{* Combinators *}


text{* The following combinators are based on the idea that two
  policies are executed in parallel. Since both input and the output
  can differ, we chose to pair them. 
  
  The new input pair will often contain repetitions, which can be
 reduced using the domain-restriction and domain-reduction
 operators. Using additional range-modifying operators such as
 $\nabla$, decide which result argument is chosen; this might be
 the first or the latter or, in case that $\beta = \gamma$, and
 $\beta$ underlies a lattice structure, the supremum or infimum of
 both, or, an arbitrary combination of them.
  
  In any case, although we have strictly speaking a pairing of
  decisions and not a nesting of them, we will apply the same
  notational conventions as for the latter, \ie{} as for flattening.
*}

text {* First the general case, then two common instances *}

definition prod_orA ::"['\<alpha>\<mapsto>'\<beta>, '\<gamma> \<mapsto>'\<delta>] \<Rightarrow> ('\<alpha>\<times>'\<gamma> \<mapsto> '\<beta>\<times>'\<delta>)" 

                       (infixr "\<Otimes>\<^sub>\<or>\<^sub>A" 55)
where "p1 \<Otimes>\<^sub>\<or>\<^sub>A p2 =
       (\<lambda>(x,y). (case p1 x of
             Some(allow d1) \<Rightarrow>(case p2 y of 
                                   Some(allow d2) \<Rightarrow> Some(allow(d1,d2))
                                 | Some(deny d2)  \<Rightarrow> Some(allow(d1,d2))
                                 | None \<Rightarrow> None) 
           | Some(deny d1)\<Rightarrow>(case p2 y of
                                   Some(allow d2) \<Rightarrow> Some(allow(d1,d2)) 
                                 | Some(deny d2)  \<Rightarrow> Some(deny (d1,d2)) 
                                 | None \<Rightarrow> None)
           | None \<Rightarrow> None))"


definition prod_orA_id:: "['\<alpha> \<mapsto> '\<beta>, '\<alpha> \<mapsto>'\<delta>] \<Rightarrow> ('\<alpha> \<mapsto> '\<beta>\<times>'\<delta>)" 
                  (infixr "\<Otimes>\<^sub>\<or>\<^sub>A\<^sub>I" 55)
where "p \<Otimes>\<^sub>\<or>\<^sub>A\<^sub>I q = (p \<Otimes>\<^sub>\<or>\<^sub>A q) o (\<lambda> x. (x,x))"


(* Probably obsolete. Todo: Check 
definition prod_orA_ST ::"['i \<times> '\<alpha> \<mapsto>  'o \<times> '\<beta>, 'i \<times> '\<gamma> \<mapsto>  'o \<times> '\<delta>] \<Rightarrow> 
                         ('i \<times> '\<alpha> \<times> '\<gamma> \<mapsto> 'o \<times> 'o \<times> '\<beta> \<times> '\<delta> )" 
                        (infixr "\<Otimes>\<^sub>\<or>\<^sub>A\<^sub>S" 55)
where "p1 \<Otimes>\<^sub>\<or>\<^sub>A\<^sub>S p2 = 
           (\<lambda> ((o1,s1),(o2,s2)). (o1,o2,s1,s2)) o_f  
           ((p1 \<Otimes>\<^sub>\<or>\<^sub>A p2) o (\<lambda>(a,b,c). ((a,b),a,c)))"
      
*)



lemma prod_orA_mt[simp]:"p \<Otimes>\<^sub>\<or>\<^sub>A \<emptyset> = \<emptyset>"
apply(rule ext, simp add: prod_orA_def)
apply (auto, simp split: option.splits decision.splits)
done

lemma mt_prod_orA[simp]:"\<emptyset> \<Otimes>\<^sub>\<or>\<^sub>A p = \<emptyset>"
by(rule ext, simp add: prod_orA_def)



lemma prod_orA_quasi_commute:
  "p2 \<Otimes>\<^sub>\<or>\<^sub>A p1 = 
      (((\<lambda>(x,y). (y,x)) o_f (p1 \<Otimes>\<^sub>\<or>\<^sub>A p2))) o (\<lambda>(a,b).(b,a))"
apply(rule ext, simp add: prod_orA_def policy_range_comp_def o_def)
apply(auto ,simp split: option.splits decision.splits)
done

definition prod_orD ::"['\<alpha> \<mapsto> '\<beta>, '\<gamma> \<mapsto>   '\<delta>] \<Rightarrow> 
                        ('\<alpha> \<times> '\<gamma> \<mapsto>  '\<beta> \<times> '\<delta> )" 
                       (infixr "\<Otimes>\<^sub>\<or>\<^sub>D" 55)
where "p1 \<Otimes>\<^sub>\<or>\<^sub>D p2 =
       (\<lambda>(x,y). (case p1 x of
             Some(allow d1) \<Rightarrow>(case p2 y of 
                                   Some(allow d2) \<Rightarrow> Some(allow(d1,d2))
                                 | Some(deny d2)  \<Rightarrow> Some(deny(d1,d2))
                                 | None \<Rightarrow> None) 
           | Some(deny d1)\<Rightarrow>(case p2 y of
                                   Some(allow d2) \<Rightarrow> Some(deny(d1,d2)) 
                                 | Some(deny d2)  \<Rightarrow> Some(deny (d1,d2)) 
                                 | None \<Rightarrow> None)
           | None \<Rightarrow> None))"


(* Probably obsolete. Check!
definition prod_orD_ST ::"['i \<times> '\<alpha> \<mapsto>  'o \<times> '\<beta>, 'i \<times> '\<gamma> \<mapsto>  'o \<times> '\<delta>] \<Rightarrow> 
                         ('i \<times> '\<alpha> \<times> '\<gamma> \<mapsto> 'o \<times> 'o \<times> '\<beta> \<times> '\<delta> )" 
                        (infixr "\<Otimes>\<^sub>\<or>\<^sub>D\<^sub>S" 55)
where "p1 \<Otimes>\<^sub>\<or>\<^sub>D\<^sub>S p2 = 
           (\<lambda> ((o1,s1),(o2,s2)). (o1,o2,s1,s2)) o_f  
           ((p1 \<Otimes>\<^sub>\<or>\<^sub>D p2) o (\<lambda>(a,b,c). ((a,b),a,c)))"
 *)


lemma prod_orD_mt[simp]:"p \<Otimes>\<^sub>\<or>\<^sub>D \<emptyset> = \<emptyset>"
apply(rule ext, simp add: prod_orD_def)
apply (auto, simp split: option.splits decision.splits)
done

lemma mt_prod_orD[simp]:"\<emptyset> \<Otimes>\<^sub>\<or>\<^sub>D p = \<emptyset>"
by(rule ext, simp add: prod_orD_def)


lemma prod_orD_quasi_commute:
" p2 \<Otimes>\<^sub>\<or>\<^sub>D p1 = 
     (((\<lambda>(x,y). (y,x)) o_f (p1 \<Otimes>\<^sub>\<or>\<^sub>D p2))) o (\<lambda>(a,b).(b,a))"
apply(rule ext, simp add: prod_orD_def policy_range_comp_def o_def)
apply (auto, simp split: option.splits decision.splits)
done

text{* The following two combinators are by definition non-commutative,
  but still strict. *}

definition prod_1 :: "['\<alpha>\<mapsto>'\<beta>, '\<gamma> \<mapsto>'\<delta>] \<Rightarrow> ('\<alpha>\<times>'\<gamma> \<mapsto> '\<beta>\<times>'\<delta>)"
                   (infixr "\<Otimes>\<^sub>1" 55)
where "p1 \<Otimes>\<^sub>1 p2 \<equiv>
       (\<lambda>(x,y). (case p1 x of
             Some(allow d1)\<Rightarrow>(case p2 y of 
                                   Some(allow d2) \<Rightarrow> Some(allow(d1,d2))
                                 | Some(deny d2)  \<Rightarrow> Some(allow(d1,d2))
                                 | None \<Rightarrow> None) 
           | Some(deny d1) \<Rightarrow>(case p2 y of
                                   Some(allow d2) \<Rightarrow> Some(deny(d1,d2))
                                 | Some(deny d2)  \<Rightarrow> Some(deny(d1,d2))
                                 | None \<Rightarrow> None)
           |None \<Rightarrow> None))"


lemma prod_1_mt[simp]:"p \<Otimes>\<^sub>1 \<emptyset> = \<emptyset>"
apply(rule ext, simp add: prod_1_def)
apply (auto, simp split: option.splits decision.splits)
done

lemma mt_prod_1[simp]:"\<emptyset> \<Otimes>\<^sub>1 p = \<emptyset>"
by(rule ext, simp add: prod_1_def)



definition prod_2 :: "['\<alpha>\<mapsto>'\<beta>, '\<gamma> \<mapsto>'\<delta>] \<Rightarrow> ('\<alpha>\<times>'\<gamma> \<mapsto> '\<beta>\<times>'\<delta>)"
                      (infixr "\<Otimes>\<^sub>2" 55)
where "p1 \<Otimes>\<^sub>2 p2 \<equiv>
       (\<lambda>(x,y). (case p1 x of
             Some (allow d1) \<Rightarrow>(case p2 y of 
                                   Some(allow d2) \<Rightarrow> Some(allow(d1,d2))
                                 | Some(deny d2)  \<Rightarrow> Some(deny (d1,d2))
                                 | None \<Rightarrow> None) 
           | Some(deny d1)\<Rightarrow>(case p2 y of
                                   Some(allow d2) \<Rightarrow> Some(allow(d1,d2))
                                 | Some(deny d2) \<Rightarrow>  Some(deny (d1,d2)) 
                                 | None \<Rightarrow> None)
           |None \<Rightarrow>None))"

lemma prod_2_mt[simp]:"p \<Otimes>\<^sub>2 \<emptyset> = \<emptyset>"
apply(rule ext, simp add: prod_2_def)
apply (auto, simp split: option.splits decision.splits)
done

lemma mt_prod_2[simp]:"\<emptyset> \<Otimes>\<^sub>2 p = \<emptyset>"
by(rule ext, simp add: prod_2_def)



lemma prod_orA_id_mt[simp]:"p \<Otimes>\<^sub>\<or>\<^sub>A\<^sub>I \<emptyset> = \<emptyset>"
apply(rule ext, simp add: prod_orA_id_def)
done

lemma mt_prod_orA_id[simp]:"\<emptyset> \<Otimes>\<^sub>\<or>\<^sub>A\<^sub>I p = \<emptyset>"
by(rule ext, simp add: prod_orA_id_def)

lemma prod_orA_id_quasi_commute:
  "p2 \<Otimes>\<^sub>\<or>\<^sub>A\<^sub>I p1 = 
           (\<lambda>(x,y). (y,x)) o_f (p1 \<Otimes>\<^sub>\<or>\<^sub>A\<^sub>I p2)"
apply(rule ext, simp add: prod_orA_id_def prod_orA_def policy_range_comp_def)
apply (simp split: option.splits decision.splits)
done


definition prod_orD_id ::"['\<alpha>\<mapsto>'\<beta>, '\<alpha>\<mapsto>'\<gamma>] \<Rightarrow> ('\<alpha> \<mapsto> '\<beta>\<times>'\<gamma>)" 
                       (infixr "\<Otimes>\<^sub>\<or>\<^sub>D\<^sub>I" 55)
where
"p \<Otimes>\<^sub>\<or>\<^sub>D\<^sub>I q = (p \<Otimes>\<^sub>\<or>\<^sub>D q) o (\<lambda>x. (x,x))"



lemma prod_orD_id_mt[simp]:"p \<Otimes>\<^sub>\<or>\<^sub>D\<^sub>I \<emptyset> = \<emptyset>"
apply(rule ext, simp add: prod_orD_id_def)
done

lemma mt_prod_orD_id[simp]:"\<emptyset> \<Otimes>\<^sub>\<or>\<^sub>D\<^sub>I p = \<emptyset>"
by(rule ext, simp add: prod_orD_id_def prod_orD_def)

lemma prod_orD_id_quasi_commute:
  "p2 \<Otimes>\<^sub>\<or>\<^sub>D\<^sub>I p1 = 
       (\<lambda>(x,y). (y,x)) o_f (p1 \<Otimes>\<^sub>\<or>\<^sub>D\<^sub>I p2)"
apply(rule ext, simp add:prod_orD_id_def prod_orD_def policy_range_comp_def)
apply (simp split: option.splits decision.splits)
done


definition prod_1_id ::"['\<alpha>\<mapsto>'\<beta>, '\<alpha>\<mapsto>'\<gamma>] \<Rightarrow> ('\<alpha> \<mapsto> '\<beta>\<times>'\<gamma>)" (infixr "\<Otimes>\<^sub>1\<^sub>I" 55)
where "p \<Otimes>\<^sub>1\<^sub>I q = (p \<Otimes>\<^sub>1 q) o (\<lambda>x. (x,x))"


lemma prod_1_id_mt[simp]:"p \<Otimes>\<^sub>1\<^sub>I \<emptyset> = \<emptyset>"
apply(rule ext, simp add: prod_1_id_def)
done

lemma mt_prod_1_id[simp]:"\<emptyset> \<Otimes>\<^sub>1\<^sub>I p = \<emptyset>"
by(rule ext, simp add: prod_1_id_def prod_1_def)


definition prod_2_id ::"['\<alpha>\<mapsto>'\<beta>, '\<alpha>\<mapsto>'\<gamma>] \<Rightarrow> ('\<alpha> \<mapsto> '\<beta>\<times>'\<gamma>)" (infixr "\<Otimes>\<^sub>2\<^sub>I" 55)
where"p \<Otimes>\<^sub>2\<^sub>I q = (p \<Otimes>\<^sub>2 q) o (\<lambda>x. (x,x))"

lemma prod_2_id_mt[simp]:"p \<Otimes>\<^sub>2\<^sub>I \<emptyset> = \<emptyset>"
apply(rule ext, simp add: prod_2_id_def)
done

lemma mt_prod_2_id[simp]:"\<emptyset> \<Otimes>\<^sub>2\<^sub>I p = \<emptyset>"
by(rule ext, simp add: prod_2_id_def prod_2_def)





text {* For constructing transition policies, two additional combinators are required: one combines state transitions by pairing the states, the other works equivalently on general maps. *}
                 
definition combine_st :: "('i \<times> '\<sigma> \<rightharpoonup> '\<sigma>) \<Rightarrow> ('i \<times> '\<sigma>' \<rightharpoonup> '\<sigma>') \<Rightarrow> 
                            ('i \<times> '\<sigma> \<times> '\<sigma>' \<rightharpoonup> '\<sigma> \<times> '\<sigma>')" (infixr "o'_st" 60) where
 "p1 o_st p2 = (\<lambda> (x,y,z). case p1 (x,y) of Some d1 \<Rightarrow>
                                     (case p2 (x,z) of Some d2 \<Rightarrow> Some (d1,d2)
                                                     | None \<Rightarrow> None)
                                                 | None \<Rightarrow> None)"

definition parallel_map :: "('\<alpha> \<rightharpoonup> '\<beta>) \<Rightarrow> ('\<delta> \<rightharpoonup> '\<gamma>) \<Rightarrow> 
                            ('\<alpha> \<times> '\<delta>  \<rightharpoonup> '\<beta> \<times> '\<gamma>)" (infixr "p'_m" 60) where
 "p1 p_m p2 = (\<lambda> (x,y). case p1 x of Some d1 \<Rightarrow>
                        (case p2 y of Some d2 \<Rightarrow> Some (d1,d2)
                                        | None \<Rightarrow> None)
                                | None \<Rightarrow> None)"


section {* Distributivity of the parallel combinators *}


lemma distr_or1_a: "(F = F1 \<Oplus> F2) \<Longrightarrow>  (((N  \<Otimes>\<^sub>1 F) o f) = 
               (((N \<Otimes>\<^sub>1 F1) o f)  \<Oplus> ((N   \<Otimes>\<^sub>1 F2)  o f))) "
apply (rule ext)
apply (simp add:  prod_1_def map_add_def 
            split: decision.splits option.splits)
apply (case_tac "f x")
apply (simp_all add: prod_1_def map_add_def 
            split: decision.splits option.splits)
done


lemma distr_or2_a: "(F = F1 \<Oplus> F2) \<Longrightarrow>  (((N  \<Otimes>\<^sub>2 F) o f) = 
               (((N \<Otimes>\<^sub>2 F1) o f)  \<Oplus> ((N  \<Otimes>\<^sub>2 F2)  o f))) "
apply (rule ext)
apply (simp add: prod_2_id_def prod_2_def map_add_def 
            split: decision.splits option.splits)
apply (case_tac "f x")
apply (simp_all add: prod_2_id_def prod_2_def map_add_def 
            split: decision.splits option.splits)
done


lemma distr_or1: "(F = F1 \<Oplus> F2) \<Longrightarrow>  ((g o_f ((N  \<Otimes>\<^sub>1 F) o f)) = 
               ((g o_f ((N \<Otimes>\<^sub>1 F1) o f)) \<Oplus>  (g o_f ((N \<Otimes>\<^sub>1 F2)  o f)))) "
apply (rule ext)+
apply (simp add: prod_1_def map_add_def policy_range_comp_def 
            split: decision.splits option.splits)
apply (case_tac "f x")
apply (simp_all add: prod_1_def map_add_def 
            split: decision.splits option.splits)
done


lemma distr_orA: "(F = F1 \<Oplus> F2) \<Longrightarrow>  ((g o_f ((N  \<Otimes>\<^sub>\<or>\<^sub>A F) o f)) = 
               ((g o_f ((N  \<Otimes>\<^sub>\<or>\<^sub>A F1) o f))  \<Oplus>  (g o_f ((N  \<Otimes>\<^sub>\<or>\<^sub>A F2)  o f)))) "
apply (rule ext)+
apply (simp add: prod_orA_def map_add_def policy_range_comp_def 
            split: decision.splits option.splits)
apply (case_tac "f x")
apply (simp_all add:  map_add_def 
            split: decision.splits option.splits)
done

lemma distr_orD: "(F = F1 \<Oplus> F2) \<Longrightarrow>  ((g o_f ((N  \<Otimes>\<^sub>\<or>\<^sub>D F) o f)) = 
               ((g o_f ((N \<Otimes>\<^sub>\<or>\<^sub>D F1) o f))  \<Oplus>  (g o_f ((N \<Otimes>\<^sub>\<or>\<^sub>D F2)  o f)))) "
apply (rule ext)+
apply (simp add: prod_orD_def map_add_def policy_range_comp_def 
            split: decision.splits option.splits)
apply (case_tac "f x")
apply (simp_all add:  map_add_def 
            split: decision.splits option.splits)
done


lemma distr_or2: "(F = F1 \<Oplus> F2) \<Longrightarrow>  ((r o_f ((N  \<Otimes>\<^sub>2 F) o f)) = 
               ((r o_f ((N \<Otimes>\<^sub>2 F1) o f))  \<Oplus> (r o_f ((N  \<Otimes>\<^sub>2 F2)  o f)))) "
apply (rule ext)
apply (simp add: prod_2_id_def prod_2_def map_add_def policy_range_comp_def
            split: decision.splits option.splits)
apply (case_tac "f x")
apply (simp_all add: prod_2_def map_add_def 
            split: decision.splits option.splits)
done

lemma coerc_assoc: "r o_f P o d = r o_f (P o d)"
apply (simp add: policy_range_comp_def)
apply (rule ext)
apply (simp split: option.splits decision.splits)
done


end

