(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * TestGen.ML  --- test data/case generation routines 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2003-2010 ETH Zurich, Switzerland
 *               2003-2010 Achim D. Brucker, Germany
 *               2003 2010 University Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: TestGen.ML 8965 2010-06-10 15:54:16Z krieger $ *)

(* Module Test *)

infix 1 |$>

infix 6 DOWNTO

infix 0 COLLECT_SS
     
signature TESTGEN =
sig
   val  mt_testenv        : TestEnv.testenv

   val trace              : bool Config.T
   (* switches on debug information *)

   val profiling          : bool Config.T
   (* switches on timer profiler *)

   val is_thyp : term -> bool
   (* Core data structure for the test environment.
      Contains tables, parameters, all sorts of "screws"
      to control the test generation process. *)

   val |$> : thm * tactic -> thm
   val TRY' : (int -> tactic) -> int -> tactic
   val COND' : (int -> thm -> bool) -> (int -> tactic) -> (int -> tactic) -> int -> tactic
   val DOWNTO : int * int -> (int -> tactic) -> tactic
   val DEPTH_SPLIT : (int -> tactic) -> int -> tactic

   val SOLVE_ASMS : Proof.context ->
        ({prems: thm list, context: Proof.context} -> tactic) -> int -> tactic

   val ALLCASES : (int -> tactic) -> tactic

   val auto_solver : thm list -> {prems: thm list, context: Proof.context} -> tactic


   val mp_fy : int -> tactic
   val all_ify : Proof.context -> string list -> int -> tactic
   val thyp_ify : Proof.context -> int -> tactic

   val var_hyp_subst_tac : int -> tactic

   val gen_test_case_tac : Proof.context -> string  -> tactic
   (* "gen_test_case_tac ctxt progname" is a 
      tactic that converts all subgoals of current proof state into 
      test-normal form - i.e. a conjunctive normal form
      (instead of DNF), where the conclusions have the
      form "P(progname x1 .. xn)", where progname does
      not occur in the premisses (the "constraints") 
      of the test. As a prerequisite, a data-separation
      theorem of depth k is resolved for up to >breadth< number
      of free variables in the current proof-state.

      NOTE: At present, the generation of data-separation
      lemmas is limited to "real" free variables and will
      not work for parameters of a test specification. Further,
      the proof-state is assumed to have only one goal.
    *)

   val gen_test_data  :  string -> Context.generic -> Context.generic
   (* "gen_test_data name context" constructs ground type instances of
      test theorems associated to "name". In a further step, 
      "abstract" or local test data were resolved into the resulting test 
      theorems (if available in the testenv under "name"), remaining
      meta variables were instantiated by a "random_solver"
      that attempts to generate an instance and tries to solve
      the constraints of a goal up to "iterations" times.
  
      The resulting simplified and "solved" test data theorems
      were stored again in the test environment.
    *)

   val get_test_data      : TestEnv.testenv * string  -> thm list
   (* get_test_data (te,name) extracts the test judgements
      (i.e. the premisses of test data statement, that are not
      test hypothesis.). *)

   val get_test_hyps      : TestEnv.testenv * string -> thm list
   (* get_test_data (te,name) extracts the test hypothesis
      of a !!test data statement!!, not a test theorem. *)

   val animate            : theory -> string * TestEnv.testenv -> unit 
   (* ground test data statements may be animated, if "P(progname x1 .. xn)"
      is executable. "animate" presents test data statements by 
      unfolding "progname x1 .. xn" and simplifying on a stepwise
      basis. (* for the moment, it's just asm_simp_tac with the
      standard evaluation trace, in the future its more or less
      the same except with slightly improved output *)
    *)

    val gen_testscript: string -> string -> string -> string -> theory -> theory


    val setup: theory -> theory
end
 
structure TestGen : TESTGEN =
struct

open HOLogic;
open TestEnv;
open Clocks;

val (trace, trace_setup) = Attrib.config_bool "testgen_trace" false;
val (profiling, profiling_setup) = Attrib.config_bool "testgen_profiling" false;


(* ************************************************************************* *)
(* access to datatype package ...                                            *)
(* ************************************************************************* *)

fun isDataType thy (Type(tn,_)) = is_some (Datatype.get_info thy tn)
   |isDataType thy _ = false;


fun get_exhaust_of type_name thy = #exhaust (Datatype.the_info thy type_name);

fun get_nchotomy_of type_name thy = #nchotomy (Datatype.the_info thy type_name);


(* ************************************************************************* *)
(* term computations ...                                                     *)
(* ************************************************************************* *)

fun max_depth (Free h) = (0,[h])
   |max_depth (Const h)= (0,[])
   |max_depth (Abs(s,tt,t)) = max_depth t
   |max_depth (s $ t) = let val (k,S) = max_depth (s)
                            val (k',S') = max_depth (t)
                        in  if k > k'+1 then (k,S)
                            else if k = k'+1 then (k, S @ S')
                                 else (k'+1, S')
                        end
   |max_depth (_) = (0,[]);


(* relies on definitions in Testing.thy ... *)
fun is_thyp (Const (@{const_name Trueprop},_) $ (Const (@{const_name THYP},_) $ _)) = true
   |is_thyp _   = false;


fun is_fun_Type (Type(@{type_name fun},_)) = true
   |is_fun_Type _               = false;



fun calc_constr_list tgt descr = 
let fun DtTFree_tvar x = (DatatypeAux.dest_DtTFree x, HOLogic.typeS)
    val sorts          = maps (fn (_,(_,s,_)) => map DtTFree_tvar s) descr
    val recTs          = DatatypeAux.get_rec_types descr sorts;
    val used           = List.foldr OldTerm.add_typ_tfree_names [] recTs;
    val newTs          = Library.take (length (descr), recTs);
    val (_,(_,insts,tgt_constrs)) = hd(filter (fn (_,(n,_,_)) => n = tgt) descr)
    val T              = hd(filter (fn (Type(n,_)) => n = tgt) newTs)
    val typ_of_dtyp    = DatatypeAux.typ_of_dtyp descr sorts
    val constr_decls   = map (fn (cname, cargs) =>
                                 Const(cname, map typ_of_dtyp cargs ---> T))
                             (tgt_constrs)
in  (map DatatypeAux.dest_DtTFree insts, constr_decls) end;



fun random_term' w n max thy cod_term_tab (Type(s,S)) =
(* w => the weight on the actual level, initial value 1
   n => level counter, inital value 0 
   max => maximal allowed number of levels
*)
       (case Symtab.lookup cod_term_tab s of
         NONE   => (* default random term generator ... 
                      TODO : should do also something for functions ... *)
                   (case Type(s,S) of 
		      Type(@{type_name int},_) => mk_number intT (IntInf.fromInt((random_range 0 20) - 10))
                    (* FIXME redundant!? -- since Isabelle2008 'a set merely abbreviates 'a => bool *)
                    | Type("set",_) => Const(@{const_name set},dummyT) $
                                    (random_term' w n max thy cod_term_tab (Type(@{type_name list},S)))
                    | Type(@{type_name fun},[T, Type(@{type_name bool}, [])]) => 
		      Const(@{const_name set}, Type(@{type_name fun}, [Type(@{type_name list}, [T]), Type(s,S)])) $
                           (random_term' w n max thy cod_term_tab (Type(@{type_name list},[T])))
                    | Type(@{type_name fun},[T,U]) => absdummy (T,(random_term' w n max thy cod_term_tab U))
                    | _     => let 
		    	           val descr = #descr(Datatype.the_info thy s)

		    	           val (insts,constrs) = calc_constr_list s descr
				   
				   val weighed_constrs = let
				       fun ar args = (length 
							      (filter (fn t => 
									  case t of 
									      DatatypeAux.DtRec _ => true 
									    | _ => false ) args) ) 
				       val constr_arity_list =  map (fn (f,args) => (f,(ar args))) 
								  (maps (#3 o snd) descr)
				   in
				       map (fn (f,a) => if a = 0 then (1,f) else (a * w,f)) 
					   constr_arity_list
				   end
			      
				   val  weighed_constrs = if (n >= max) 
							  then filter (fn (w,f) => w =1) weighed_constrs
							  else weighed_constrs
			           fun weight_of t = fst(hd ((filter (fn (w,ty) => ty=t)) weighed_constrs)) 

				   fun frequency' xs = let 
				       val max = List.foldl Int.max 0 (map fst xs);
				       val xs' = map (fn (x,a) => (max-x+1,a)) xs
				   in
				       frequency xs'
				   end
				   (* the default is a random bias towards constants *)
				   val constr = frequency' weighed_constrs
                                   val Const(h,ty) = hd (filter (fn Const(h,ty) => h = constr) constrs) 
				   val w          = weight_of h 
		    	           val ty_binds   = insts ~~ S 
		    	           fun ty_inst (s,_) = the (AList.lookup (op =) ty_binds s)
		    	           val instantiated_ty = map_type_tfree ty_inst ty
		    	           val const_head = Const(h,instantiated_ty)
		    	           val arg_ty = binder_types instantiated_ty
		    	       in  list_comb(const_head,
                                             (map(random_term' w (n+1) max thy cod_term_tab)(arg_ty)))
                               end (* FIXME avoid handle _ *)
                               handle _=>error("Cannot generate random value for type:" ^s))
       | SOME R => R S)
   |random_term' _ _ _ _ _ _ = error "Internal error in random_term: type not ground";


fun random_term thy cod_term_tab typ = random_term' 1 0 10000 thy cod_term_tab typ  

(* test section:

val ttt = [HOLogic.intT,HOLogic.unitT,HOLogic.boolT,
           HOLogic.mk_setT HOLogic.intT, 
           HOLogic.listT HOLogic.intT];
map (random_term Symtab.empty) ttt;

 *)


(* complex type instantiations ... *)

fun groundT_thm thy bound candTs thm =
    let val thm   = Thm.varifyT thm
        val tvars = OldTerm.term_tvars (list_comb(concl_of thm,prems_of thm))
        fun ctyp_of_sort x so t = if Sign.of_sort thy (t,so) 
                                  then (SOME(ctyp_of thy (TVar(x,so)), ctyp_of thy t) 
				             handle _ => NONE)  (* FIXME avoid handle _ *)
                                  else NONE
        fun prodS [] B = []
           |prodS (a::R) B =  (map(fn x => a::x) B) @ (prodS R B)

        fun compute_instsT [] = [[]]
           |compute_instsT ((x,so)::R) = prodS (List.mapPartial (ctyp_of_sort x so) candTs)
                                               (compute_instsT R)
    in 
	map (fn x => instantiate (x,[]) thm)
	    (Library.take(bound,compute_instsT tvars))
    end;
    
(* test

 val thm' = groundT_thm thy
                        (#type_range_bound(rep_testenv te)) 
                        (#type_candicates(rep_testenv te)) 
                        thm

 *) 

(* *********************************************************************** *)
(*                                                                         *)
(* SOME more universal tacticals -> tctical.ML                             *)
(*                                                                         *)
(* *********************************************************************** *)

fun thm |$> tac = 
    case Seq.pull (tac thm) of
         SOME (m,_) => m
        |_ => error"|$> : TACTIC FAILED" 

fun TRY' tac n = TRY(tac n) (* standard combinator,
                               not implemented so far in Isabelle *)


fun COND' testfun thenf elsef x prf = 
      if testfun x prf then thenf x prf else elsef x prf;


fun (k DOWNTO (n:int)) = fn tac => (fn st =>
  let fun doall x = if x < n 
                    then all_tac 
                    else ((tac x) THEN doall(x-1))
      val m = nprems_of st
  in  if k <= 0 orelse k > m
      then all_tac st
      else doall k st
  end)


fun DEPTH_SPLIT tac n st = 
(* PRE tac should only exhausting - i.e. increase
       the number of subgoals finitely many times
   POST if tac never fails, DEPTH_SPLIT tac will not fail.
 *)
    let val m   = nprems_of st
        val res = st |$> (tac n)
        val m'  = nprems_of res
    in  if m' > m
        then ((n + m' - m) DOWNTO n)(DEPTH_SPLIT tac) (res) 
        else all_tac res
    end 
         


fun SOLVE_ASMS ctxt tac n state =
    let val goal = Logic.get_goal (prop_of state) n
        val params = Logic.strip_params goal
        val prems =  map HOLogic.dest_Trueprop (rev(Logic.strip_imp_prems goal))
        val eqTrueI2 = @{lemma "[|P;P |] ==> True" by auto}
        fun decomp thm n = let val mm = (thm RS conjunct1)
                           in  (dtac (mm RS eqTrueI2) n) THEN
                               (etac @{thm TrueE} n) THEN
                               (decomp (thm RS conjunct2) n) end
                           handle THM _ => (dtac (thm RS eqTrueI2) n) THEN
                                           (etac @{thm TrueE} n)

    in  case prems of
        [] => all_tac state
       |a::r => let val prem = List.foldl HOLogic.mk_conj a r
                   val prem'= (Term.subst_bounds ((rev(map Free params)),prem))
                   val thm  = Goal.prove ctxt [] [] (HOLogic.mk_Trueprop prem') tac
                in (decomp thm n) state end
                handle _ => no_tac state (* cold not solve asms ... *)  (* FIXME avoid handle _ *)
    end


fun auto_solver thms {prems: Thm.thm list, context: Proof.context} state =
    let val css   = claset_of context
        val ss    = simpset_of context
        val thm'  = state |$> (TRY(safe_tac (css delrules [notI])))
        val goals = Thm.nprems_of thm'
        fun solve_thm thm = FIRSTGOAL (fn n => (rtac thm n) THEN 
                                       (IF_UNSOLVED (SOLVE(auto_tac (css,ss)))))
        fun solve_thms [] = no_tac
           |solve_thms (a::R) = solve_thm a ORELSE solve_thms R
    in  if goals > 0 then (solve_thms thms) thm' else all_tac thm'
    end
    

(* *********************************************************************** *)
(*                                                                         *)
(* SOME Testing specific tacticals                                         *)
(*                                                                         *)
(* *********************************************************************** *)

(* PRE n>0 and n <= nprems_of thm *)
fun is_THYP n thm = is_thyp(Logic.nth_prem(n, prop_of thm))


fun ONLY_CASES tac = (COND' is_THYP (K all_tac) tac)


fun ALLCASES_before k tac = (k DOWNTO 1) (ONLY_CASES tac)


fun ALLCASES tac thm = (ALLCASES_before (nprems_of thm) tac thm)


fun unfold rules = asm_full_simp_tac (HOL_basic_ss addsimps rules)

(* Optimization potential: safe_tac against something more
   specialized should do ... *)
fun if_splitter cs  = asm_full_simp_tac(HOL_basic_ss addsplits if_splits)
                   THEN' TRY' (fn n => SELECT_GOAL (safe_tac cs) n)


(* *********************************************************************** *)
(*                                                                         *)
(* Initial test environment                                                *)
(*                                                                         *)
(* *********************************************************************** *)

val _ = Context.>> (TestEnv.map_data
  (pre_safe_tac_update (fn ctxt => ALLCASES (asm_full_simp_tac (simpset_of ctxt)))));

val mt_testenv = get_data (ML_Context.the_local_context ());



(* *********************************************************************** *)
(*                                                                         *)
(* basic tactics ...                                                       *)
(*                                                                         *)
(* *********************************************************************** *)

fun mp_fy n = (REPEAT(etac rev_mp n));

fun all_ify ctxt excpt n state = 
    let val subgoal   = List.nth(prems_of state,n-1)
        val params    = Logic.strip_params subgoal
        val frees     = map dest_Free (OldTerm.term_frees subgoal)
        fun sbs(n,t)  = res_inst_tac ctxt [(("x",0), n)] spec
        val vars      = filter (fn x => not((fst x) mem excpt))
                                   (params@frees)
    in  (if null vars then no_tac 
         else List.foldr(fn(x,ta) => (sbs x n) THEN ta)
                   prune_params_tac vars) 
        (state) 
    end

fun thyp_ify ctxt n =
    EVERY[mp_fy n, all_ify ctxt [] n,
          rtac (@{thm THYP_def} RS @{thm meta_eq_to_obj_eq} RS @{thm iffD1}) n]




fun convert_goals_to_metahyps (test:(term -> bool)) thm = 
    let val thy = theory_of_thm thm
    in  map (fn prop => Thm.forall_elim_vars 0 (assume (cterm_of thy prop)))
            (filter test (prems_of thm))
    end;


fun instantiate_tac insts n thm =
    let val goal = cterm_of (theory_of_thm thm) (List.nth(prems_of thm,n-1))
    in  (rtac(instantiate ([],insts)(trivial goal)) n) thm end


fun var_hyp_subst_tac no state =
    let val prms = Logic.strip_imp_prems(List.nth(prems_of state,no-1));
        fun is_eq (Const (@{const_name "op ="}, _) $ (Var _) $ _) = true
           |is_eq (Const (@{const_name "op ="}, _) $ _ $ (Var _)) = true
           |is_eq _ = false
        val i = Library.find_index(fn (Const(@{const_name Trueprop},_)$X) => is_eq X) 
                                  (prms)
        fun mks thm = if i >= 0 
                      then ((rotate_tac i no) THEN etac @{thm thin_refl} no) thm
                      else no_tac thm
    in ((REPEAT mks) 
        THEN 
        (full_simp_tac (global_simpset_of @{theory Main}) no)) state
    end;

(* *********************************************************************** *)
(*                                                                         *)
(*  special tactics for TestGen                                            *)
(*                                                                         *)
(* *********************************************************************** *)


fun mk_Uniformity_THYP term =
    let fun convAll(Const(@{const_name All},ty)$Abs(n,ty',tt)) = 
                      (Const(@{const_name Ex},ty)$Abs(n,ty',convAll tt))
           |convAll tt = tt;
        val tt   = dest_Trueprop term
        val thyp = Const (@{const_name THYP}, [boolT]--->boolT) $
                        (Const (@{const_name "op -->"}, [boolT,boolT]--->boolT)
                         $ (convAll tt) $ tt)
    in  mk_Trueprop thyp end

fun add_term_consts (Const (c, _), cs) = insert (op =) c cs
  | add_term_consts (t $ u, cs) = add_term_consts (t, add_term_consts (u, cs))
  | add_term_consts (Abs (_, _, t), cs) = add_term_consts (t, cs)
  | add_term_consts (_, cs) = cs;

fun term_consts t = add_term_consts(t,[]);

fun unifomityI_tac ctxt prog n state =
    let fun prem_of state = List.nth(prems_of state, n-1)
        fun core_tac state = subgoal_terminst_tac [] 
                                (mk_Uniformity_THYP (prem_of state)) 
                                (n) (state)
	val count = Unsynchronized.ref 0;
	fun nwstr() = (count:=(!count)+1; "?X"^(Int.toString(!count))^"X")
    in (if "TestGen.THYP" mem (term_consts(prem_of state))
        then no_tac
        else  EVERY[mp_fy n, all_ify ctxt [prog] n, 
                    core_tac, etac @{thm THYP_app1} n, 
                    REPEAT(fn t => (res_inst_tac ctxt [(("x",0),nwstr())] exI n)(t)), 
		    strip_tac n]) 
       state
    end


(* lemma *)
fun swap_accumulate thy = let
    val prem_terms = [@{term "P"}, @{term "~P | Q"}]
    val prems = map (assume o (cterm_of thy)) prem_terms
in
    (trivial (cterm_of thy @{term "[| P ; ~P | Q |] ==> Q"})) 
	|$> (cut_facts_tac prems 1) 
	|$> (auto_tac (global_claset_of thy, global_simpset_of thy))
end


(* ************************************************************************* *)
(* data separation lemma generation ...                                      *)
(* ************************************************************************* *)

fun refine_data_separation_lemma changed ctxt bal prem_no cands thm =
(* - bal controls of the refinement is in balancing mode
     (only parameters whose occurrence in the term is
      has strictly less than maximum height)
     or not. 
   - prem_no indicates the number of the premise
   - cands is the global list of candidates.
     In balancing mode, parameters in maximal depth
     position were eliminated.
   - PRE maxidx is 0 for all mvars
   - POST maxidx is 0 for all mvars

 *)
let 
    val thy    = ProofContext.theory_of ctxt
    val prem   = Logic.nth_prem(prem_no, prop_of thm) (* besser: effizienter!*)
    val params = Logic.strip_params(prem)
    val cands' = if bal then
                    let val term   = HOLogic.dest_Trueprop(hd
                                            (Logic.strip_assums_hyp prem))
                        val (k,mp) = max_depth(Term.subst_bounds
                                            ((rev(map Free params)),term))
                    in filter(fn p => forall(fn p'=> not(p=p')) (mp)) cands end
                 else cands;

    val var_no = find_index (fn(s,t)=> exists(fn (name,ty)=>
                                             s = name andalso t = ty) cands')
                             params
    val (name,ty)  = if var_no > ~1 then List.nth(params,var_no)
                     else raise (ERROR("no param found"))
    val Type(tn,_) = ty;
    val m          = get_exhaust_of tn thy;
    val n          = m RSN (prem_no, thm);
    (* pre: all occurrences of ?indname t1 ... tn have the same
            arguments and types  and occur in a critical pattern : 
             ?indname t1 ... tn = C x1 ... xn 
        ... 
       The Thm.adjust_maxidx_thm reflects the fact, that only one
       MVar y should have been left - the others are erased due to
       hypsubst internally
    *)
    fun find_mvar_pat indname goal_no thm =
        let val goal = Logic.get_goal thm goal_no
            val params = Logic.strip_params goal
            fun  cc(Const(@{const_name "==>"}, _) 
                    $ (Const(@{const_name Trueprop}, _) 
                       $ (Const(@{const_name "op ="},t) $ A $ B)) 
                    $ C) = (case strip_comb(subst_bounds(
                                     rev(map Free params), A)) of
                             (Var(s,_),S) =>(if s = indname 
                                             then (map dest_Free S, head_of B)
                                             else cc(C))
                            | (_, _)  => cc C)
               | cc _ = error("find_mvar_pat: No pattern found")        
        in cc (strip_all_body goal) end
    fun lifter name S =
        let fun ll k [] = if 0 <=k then Bound(k) 
                          else error"find_mvar:impossible lifter"
               |ll k ((s,t)::R) = 
                      if s = name then Abs(s,t, ll 0 R)
                      else Abs(s,t, ll (k+1) R)
        in ll (~(length S)) S end
    fun typ_of_lifter (name:string) S = 
        let fun tll S [] = error"find_mvar:impossible type for lifter"
               |tll S ((s,t)::R) = 
                       if name = s 
                       then ((rev S)@(t::(map snd R)), t)
                       else tll (t::S) R
        in  (op --->)(tll [] S) end   
    val (lifter_params, constr) = find_mvar_pat ("y",1) prem_no (prop_of n)
    (* ("y",1) is a fact from the invariant maxidx=0 (see below) *)
    val subst0 = lifter name lifter_params 
    val ty_lift = fastype_of subst0
    val ty_constr = fastype_of constr
    val subst = cterm_of thy subst0 
    val ty_lift_body = body_type ty_lift
    val ty_lift_constr = body_type ty_constr
    val tyenv = Sign.typ_match thy (ty_lift_constr,ty_lift_body) Vartab.empty
                handle Type.TYPE_MATCH => Vartab.empty
    (* Hack that works for cases, where non-recursive positions in a type
       were GROUND types (like color in RBT_tree). *)  
    fun tysubs2ctyps(x,(s,t)) = (ctyp_of thy (TVar(x,s)),ctyp_of thy t)
    val tsubst = map tysubs2ctyps (Vartab.dest tyenv)
    val ()    = (changed:=true)
    val var = cterm_of thy (Var(("y",1),ty_lift))
    val result = (instantiate (tsubst, [(var,subst)]) n)
                 |$> (ALLGOALS(fn n => TRY(hyp_subst_tac n))) 
                 |$> prune_params_tac
in  Thm.adjust_maxidx_thm 0 result 
    (* establishes POST and makes this fact therefore invariant. *) 
    (* More safe, but also more expensive variant: standard result. *)
end

(* This is the core procedure of the data_separation_lemma generation, and,
   in a way, the core of the test case generator itself.

   It proceeds as follows: in a given theorem thm (usually an exhaustion lemma 
   itself), the premisses (having the form !!x1..xn. A x1 .. xn ==> B in 
   general) were selected and "refined" in a "sweep". This means that in a 
   sweep, for each premisse an xi is selected (which must have a data-type, 
   say ti) and the exhaustion theorem of ti is forward resolved into this 
   premisse; cleaning it up (by prems_hyp_subst_tac) also erases xi as 
   parameter of this premisse. 
   In a next sweep, another xj may be selected and so forth, up to the moment 
   were no parameters remained in the premisse that existed before the initial 
   sweep (this the list parameters is the "cand" (candidates) x1 .. xj for 
   all premisses which precomputed.)

   Then, the candidate list may be recomputed and the process repeated up to 
   n times. Since any group of sweeps erases one generation of parameters, 
   the parameter n corresponds to "depth+1" of the data term occuring in the 
   premisse increases, i.e. from:

   [| ?y = Leaf ==> ?P; 
      !!a tree1 tree2. ?y = Node a tree1 tree2 ==> ?P |] 
   ==> ?P

   we get:

   [| ?y = Leaf ==> ?P; !!a. ?y = Node a Leaf Leaf ==> ?P;
      !!a aa tree1 tree2a. ?y = Node a Leaf (Node aa tree1 tree2a) ==> ?P;
      !!a aa tree1a tree2a. ?y = Node a (Node aa tree1a tree2a) Leaf ==> ?P;
      !!a aa tree1a tree2a ab tree1 tree2b.
            ?y = Node a (Node aa tree1a tree2a) (Node ab tree1 tree2b) ==> ?P 
   |] ==> ?P

   This picture is slightly complicated by the fact that Isabelle does not
   change the parameters, but may reintroduce parameter erased by a previous 
   sweep during the forward resolution step. Thus, the procedure may loop.

   Instead of not erasing the parameters during lemma-refinement (this
   leads to very large liftings of the meta-variables over these
   parameter lists and is thus expensive), we chose to use 
   a particular test for "maximal" parameters in order to avoid this problem.

 *)  
fun nrefine_data_separation_lemma changed ctxt 0 thm = thm
  | nrefine_data_separation_lemma changed ctxt n thm =
    let 
      val thy = ProofContext.theory_of ctxt
      val isaDataType   = fn(_,ty)=> isDataType thy ty
      val thm_params    = maps Logic.strip_params (prems_of thm)
      val cands         = distinct (op=) (filter isaDataType thm_params)
      fun rds b(prno,t) = refine_data_separation_lemma changed ctxt b prno cands t 
                          handle _ => t  (* FIXME avoid handle _ *)
      val ()            = (changed:=false)
      fun sweep b th    = List.foldr (rds b) th (1 upto (length(prems_of th)))
      val thm1          = sweep false thm       (* debalancing phase *)
      fun repeat_till_stable thm = let val ()   = (changed:=false)
                                     val thm' = sweep true thm
                                   in  if !changed 
                                       then repeat_till_stable thm' 
                                       else thm'
                                   end
    in  nrefine_data_separation_lemma changed ctxt
          (n-1) 
           (if  !changed 
            then thm1                     (* debalancing phase had no effect *)
            else repeat_till_stable thm1) (* balancing phase *)
    end;

(* NOTE: Currently, this core routine is restricted to
   "real free" variables; parameters will not be
   considered as a simple means to avoid looping ... *)
fun data_separation_tac ctxt depth no state =
    let 
        val thy    = ProofContext.theory_of ctxt
        val prem   = List.nth(prems_of state, no - 1)
        val params = Logic.strip_params(prem)
        val term   = HOLogic.dest_Trueprop(Logic.strip_assums_concl prem)
        val frees  = map dest_Free (OldTerm.term_frees term)
(*      val cands  = filter (fn (_,ty) => isDataType thy ty) (params@frees) *)
        val cands  = filter (fn (_,ty) => isDataType thy ty) (frees)
        fun ds_thm tn =
          nrefine_data_separation_lemma (Unsynchronized.ref false) ctxt depth (get_exhaust_of tn thy)
    in (case cands of 
          [] => no_tac
        | (na,Type(tn,_))::_ => 
                    let val data_sep_thm = Drule.standard(ds_thm tn)
                        val brchs = length(prems_of data_sep_thm)
                    in       (res_inst_tac ctxt [(("y",0),na)] data_sep_thm no)
                        THEN (res_inst_tac ctxt [(("x",0),na),
                                                   (("k",0),Int.toString depth)] 
                                                  @{thm THYP_E_reg}
                                                  (no + brchs - 1))
                        THEN (thin_tac ctxt (na^"= ?X") (no+brchs))            
                    end)
       (state)
    end;




(* ************************************************************************* *)
(* normal form computation ...                                               *)
(* ************************************************************************* *)

fun to_TNF ctxt prog term no state =
    let
	val thy     = ProofContext.theory_of ctxt;
	val prems   = Logic.prems_of_goal term no
	val concl   = Logic.concl_of_goal term no

        val prems'  = filter (fn t => prog mem (OldTerm.add_term_names(t,[]))) prems;
        val concl'  = filter (fn t => prog mem (OldTerm.add_term_names(t,[]))) [concl];
	fun subst t = (Thm.cterm_of thy (Var(("P",0),HOLogic.boolT)),
                       Thm.cterm_of thy t);
        (*  swap_tac : tactic * term -> tactic; takes one term containing 
            "prog" and swap it into the conclusion as disjoint *)
        fun swap_tac (tm,tc) = tc THEN 
                               (eres_terminst_tac 
                                  []
                                  [subst (HOLogic.dest_Trueprop tm)]
                                  (swap_accumulate thy)
                                  no)
        (* rev_tac swaps term "a" into conclusion *)
        fun rev_tac thy a = eres_terminst_tac 
	                          [] 
                                  [subst (HOLogic.dest_Trueprop a)] 
                                  @{thm rev_notE} 
                                  no
    in ((case (prems',concl') of
          ([],[])   => error ("Subgoal " ^Int.toString(no)
                             ^" can not be converted into TNF.")
          (* case that conclusion does not contain "prog", but several
             assumptions do: pick one of them, revert first *)
         |(a::R,[]) =>(rev_tac thy a
                       THEN
                       (List.foldl swap_tac all_tac R))
         |(R,_)     => (List.foldl swap_tac all_tac R)
        )
        THEN
         (simp_tac HOL_ss no)) state
    end;


(* TODO: adb *)
(*
fun normalize_TNF prog top =
    let val prems = prems_of top
        val mrs   = prems ~~ (1 upto (length prems))  
    in  List.foldl (fn(arg,t)=> t THEN to_TNF prog arg) all_tac mrs (top)
    end
 *)
fun normalize_TNF ctxt prog top =
    let val mrs   = (length (prems_of top)) downto 1
    in  List.foldl (fn (arg,t) => t THEN to_TNF ctxt prog (prop_of top) arg) 
              all_tac 
	      mrs 
	      top
    end


(* TODO: adb *)
fun minimize_TNF ctxt = distinct_subgoals_tac
(* This should do:
   - sorting assumptions in all subgoals
   - sorting parameters of subgoals
   - sortieren der clauses - 
     weniger asms zuerst (defer_tac)
   - eliminieren von assumptions subsumed in other clauses
     [| P;R |] ==> A;
     [| P;Q;R  |] ==> A
     ------------------
     [| P;R |] ==> A;
     [| P;R  |] ==> A
     durch thin_tac
   - eliminieren von subsumed or's,
     a particular weakness of our procedure so far.
     [| P;R |] ==> A;
     [| ~P;Q |] ==> B
     [| R; Q|] ==> A|B;
     ------------------
     [| P;R |] ==> A;
     [| ~P;Q |] ==> B
    - finally distinct_subgoals_tac
*)


(* orig version *)
(*
fun gen_test_case_tac thy clasimp depth breadth prog  =
    EVERY[ALLGOALS (asm_full_simp_tac (snd clasimp)),
          safe_tac ((global_claset_of thy) addss (global_simpset_of thy)),
          normalize_TNF thy prog,
          minimize_TNF,
          ALLGOALS (fn n => TRY(unifomityI_tac prog n))];
*)



(* ************************************************************************* *)
(* test case generation ...                                                  *)
(* ************************************************************************* *)


fun gen_test_case_tac ctxt prog  =
    let val testenv = TestEnv.get_data ctxt
        val depth = Config.get ctxt depth
        val breadth = Config.get ctxt breadth

        val prep_HCNF = get_pre_safe_tac testenv
        val to_HCNF = TRY (safe_tac (claset_of ctxt))
        val pre_normalize_TNF_tac = get_pre_normalizeTNF_tac testenv 
        val pre_minimize_TNF_tac = get_pre_minimizeTNF_tac testenv
        fun trace_tac b str = (if Config.get ctxt trace then print_tac ("\n"^b^":"^str^"\n")
	                       else all_tac) 
        fun exec phase b str = (if Config.get ctxt profiling then start_clock_tac str else all_tac) 
                               THEN phase THEN trace_tac b str THEN
                               (if Config.get ctxt profiling then stop_clock_tac str else all_tac)

        fun standard k b = 
            EVERY[if Config.get ctxt trace then print_tac "Enter" else all_tac,
                  exec (prep_HCNF ctxt) b "Simp",
                  exec (to_HCNF) b "HCN",    
                  exec (pre_normalize_TNF_tac ctxt) b "pre_norm",
                  exec (normalize_TNF ctxt prog) b "TNF",                           
                  exec (pre_minimize_TNF_tac ctxt) b "pre_minimize",        
                  exec (minimize_TNF ctxt) b "MinimTNF"]
        val dst = data_separation_tac ctxt depth  


	(*True if the two theorems have the same prop field, ignoring hyps, der, etc.*)
	val eq_thm_prop = op aconv o pairself Thm.full_prop_of;

        fun main k st =(if k < breadth (* not yet all frees per goal 
                                          less breadth exhausted*)
                        then   (ALLCASES(fn n => TRY(dst n) )) THEN
                               (* try exhaustion per goal *)
                               (COND(fn st' => not (eq_thm_prop(st, st')))
                                        (* something 's changed *)
                                        (     (ALLCASES (fn n => TRY(hyp_subst_tac n) ) ) (* REALLY ? *)
                                         THEN (standard k "1")
                                         THEN (main (k+1))
                                        ) 
                                        (standard k "2"))
		        else standard  k "0" (* run out of breadth *)) (st)
	val begin_profiling_tac = if Config.get ctxt profiling then 
				      start_clock_tac "unnamed_test_thm" THEN start_clock_tac "gen_test_cases" 
				  else 
				      all_tac

	val end_profiling_tac = if Config.get ctxt profiling then 
				      stop_clock_tac "gen_test_cases" THEN stop_clock_tac "unnamed_test_thm"
				  else 
				      all_tac
    in  EVERY[begin_profiling_tac,
              exec (ALLCASES((asm_full_simp_tac (simpset_of ctxt)))) "" "pre-simplification",
              exec (main 0) "" "main completed", 
              exec (ALLCASES(TRY'(unifomityI_tac ctxt prog))) "" "main uniformity NF",
              end_profiling_tac] 
    end;


(* ************************************************************************* *)
(*                                                                           *)
(* testdata generation        ...                                            *)
(*                                                                           *)
(* ************************************************************************* *)

  (* *********************************************************************** *)
  (* Random Solver                                                           *)
  (* *********************************************************************** *)

fun random_insts thy cod_tab vars = map(fn(x as Var(s,t))=>
					   (cterm_of thy x,
					    cterm_of thy (random_term thy cod_tab t))) vars

fun list_abs_var t [] = t
  | list_abs_var t ((x as Var(_, T))::vars) = Abs(Name.internal Name.uu, T, abstract_over(x, list_abs_var t vars))

fun abs_vars t = list_abs_var t (OldTerm.term_vars t);

fun iterate f 0 = NONE
  | iterate f k = case f () handle Match => NONE
		   of NONE => iterate f (k - 1) | SOME q => SOME q;

fun certify_pairs thy l = map (fn (x,y) => (cterm_of thy x, cterm_of thy y)) l

fun uncertify_pairs l = map (fn (x,y) => (term_of x, term_of y)) l

fun quickcheck_test_term ctxt vars size i t =
  let
      val thy = ProofContext.theory_of ctxt
      val neg =  @{term Not} $ t
      val neg' = list_abs_var neg vars
      val tester = Codegen.test_term ctxt neg' 
      fun with_tester k = iterate (fn () => tester k) i
      fun with_size k = if k > size then NONE
			else 
			    (case with_tester k
			      of NONE => with_size (k + 1) 
			       | SOME q => SOME q);
  in case with_size 1 of
	 SOME insts => SOME (certify_pairs thy (vars ~~ insts))
       | NONE => NONE
  end

fun check_thm ctxt thm insts = let	
    val instantiated = instantiate ([], insts) thm
    val folded = Seq.hd (full_simp_tac (simpset_of ctxt) 1 instantiated)
in
    (nprems_of folded) < (nprems_of thm)
end

fun check_insts ctxt t =
    check_thm ctxt (trivial (cterm_of (ProofContext.theory_of ctxt) (HOLogic.mk_Trueprop t)))

fun guess_insts thy checker cod_tab vars = let
    val insts = random_insts thy cod_tab vars
in
    if checker insts then
	SOME insts
    else NONE
end

fun simple_test_term ctxt cod_tab vars i t = let
    val thy = ProofContext.theory_of ctxt
    val thm = trivial (cterm_of thy (HOLogic.mk_Trueprop t))
    val checker = check_insts ctxt t
    fun tester () = guess_insts thy checker cod_tab vars
in
    iterate tester i
end

fun combined_test_term ctxt cod_tab vars t = let
    val simple_iterations = 50
    val simple_result = simple_test_term ctxt cod_tab vars simple_iterations t
    val further_iterations = Int.max ((Config.get ctxt iterations) - simple_iterations, 0)
in
    case simple_result of
	SOME insts => SOME insts
      | NONE => quickcheck_test_term ctxt vars 15 further_iterations t
        (* FIXME avoid handle _ *)
	handle _ => simple_test_term ctxt cod_tab vars further_iterations t
end
			     
(* val tc_timer =  Unsynchronized.ref (Timer.startRealTimer ()) *)
fun random_inst_tac ctxt cod_tab hint n thm = let
    val thy = ProofContext.theory_of ctxt
    val prems = prems_of thm
    val prem  = List.nth(prems,n-1)
    val ts = Logic.strip_imp_prems prem
    val k = length ts
    val ts' = map HOLogic.dest_Trueprop ts
    val t = List.foldr (fn (x,y) => @{term "op &"} $ x $ y) @{term True} ts'
in
    if (is_some hint) andalso (check_insts ctxt t (the hint)) then 
	full_simp_tac (simpset_of ctxt) n (instantiate ([], (the hint)) thm)
    else let
	    val vars = OldTerm.term_vars t
	    val allvars = OldTerm.term_vars prem
	    val _     = tracing ("Random solving subgoal "^Int.toString(n)) 
				(* val td_time = Timer.checkRealTimer (!tc_timer) *)
	    val insts = combined_test_term ctxt cod_tab vars t
					   (* val tm = Time.-(Timer.checkRealTimer (!tc_timer),td_time) *)
	    val intro_RSF = (REPEAT(thin_tac ctxt "?X" n)) 
				THEN (rtac @{thm RSF_E} n)
			    
	    val any_insts = random_insts thy cod_tab vars
	    val extra_insts = random_insts thy cod_tab (OrdList.subtract TermOrd.term_ord vars allvars)
	in
	    case insts of
		NONE => intro_RSF (instantiate ([], extra_insts @ any_insts) thm)
	      | SOME l => full_simp_tac (simpset_of ctxt) n (instantiate ([], extra_insts @ l) thm)
	end
end;

fun pretty_insts ctxt msg insts = let
    fun mkeq (x,y) = Const(@{const_name "op ="}, dummyT) $ x $ y
    val pretty_insts = map (fn inst => (Syntax.pretty_term ctxt (mkeq inst))) insts
in
    Pretty.string_of (Pretty.big_list (msg^":") pretty_insts)
end

fun read_func_def (((u $ arg ), value), deftab) = 
    if (is_Free u) andalso (OldTerm.term_frees arg) = [] then let
	    val Type(@{type_name fun}, [U,V]) = fastype_of u
	    val updT = (U --> V) --> U --> V --> U --> V
	    val f' = case Termtab.lookup deftab u of
		NONE => absdummy (U,value)
	      | SOME f => Const(@{const_name fun_upd}, updT) $ f $ arg $ value
	in 
	    Termtab.update (u, f') deftab
	end
    else
	deftab

| read_func_def (_, deftab) = deftab

fun refine_defs defs = let
    val (lhs,rhs) = split_list defs
    val remaining_unknowns = (maps OldTerm.term_frees rhs) @ (maps OldTerm.term_vars rhs)
    val remaining_unknowns' = OrdList.make TermOrd.term_ord remaining_unknowns
    fun good_inst (x,y) = (OrdList.member TermOrd.term_ord remaining_unknowns' x) 
                          andalso ((OldTerm.term_frees y) = [])
                          andalso ((OldTerm.term_vars y) = [])

    val insts = filter good_inst defs
    val subst = map (subst_atomic insts)
in
    if insts = [] then
	defs
    else
	refine_defs (lhs ~~ (subst rhs))
end

fun scan_def (Const (@{const_name "op ="}, _) $ t $ u) = (t,u)
  | scan_def _ = error ("Unexpected SMT counterexample format")
                 
fun mkinsts thy vars ce = let
    val defs = map scan_def ce
    (* val _ = tracing (pretty_insts @{context} "Raw SMT counterexample" defs) *)
    val defs' = defs @ (map swap defs)
    fun is_Free_or_Var t = (is_Free t) orelse (is_Var t)
    val non_func_defs = filter (fn (x,_) => is_Free_or_Var x) defs'
    val potential_func_defs = filter (fn (x,_) => not (is_Free_or_Var x)) defs'
    val refined_non_func_defs = refine_defs non_func_defs
    fun good_inst (_,y) = ((OldTerm.term_frees y) = []) andalso ((OldTerm.term_vars y) = [])
    val non_func_insts = filter good_inst refined_non_func_defs
    val subst = map (subst_atomic non_func_insts)
    val (lhs,rhs) = split_list potential_func_defs
    val refined_func_defs = (subst lhs) ~~ (subst rhs)
    val rewritten_func_defs = Termtab.dest (List.foldl read_func_def Termtab.empty refined_func_defs)
    val refined_defs = refine_defs (refined_non_func_defs @ rewritten_func_defs)
    val var_insts = map (fn x => (x, subst_atomic refined_defs x)) vars
    val good_insts = filter good_inst var_insts
    (* val _ = tracing (pretty_insts @{context} "Generated instantiation from SMT counterexample" good_insts) *)
in
    certify_pairs thy good_insts
end

(* a wrapper around smt_tac' that fixes the provided counterexamples *)
fun smt_ce_tac ctxt rules =
    SUBPROOF(fn {context, prems, schematics, ...} => (fn thm =>
                (Seq.single (Seq.hd (SMT_Solver.smt_tac' true context (rules @ prems) 1 thm)))
                handle SMT_Solver.SMT_COUNTEREXAMPLE (b,ce) => let
                           val inv_insts = map swap (snd schematics)
                           val term_insts = uncertify_pairs inv_insts
                           val ce' = map (subst_atomic term_insts) ce
                       in
                           raise SMT_Solver.SMT_COUNTEREXAMPLE (b,ce')
                       end)) ctxt

fun add_bound_tac ctxt bound t = 
    Subgoal.FOCUS(fn {context, schematics, ...} => let
                         val thy = ProofContext.theory_of context
                         val ct = instantiate_cterm schematics (cterm_of thy t)
                         val bound_term = cterm_of thy (HOLogic.mk_nat bound) 
                         val xtype = ctyp_of thy (TVar (("'a",0), @{sort "Nat.size"}))
                         val tinst = (xtype, ctyp_of_term ct)
                         val k = cterm_of thy (Var (("k",0), @{typ "nat"}))
                         val x = cterm_of thy (Var (("x",0), type_of t))
                         val inst1 = (k, bound_term)
                         val inst2 = (x, ct)
                     in
                         res_terminst_tac [tinst] [inst1, inst2] @{thm add_bound} 1
                     end) ctxt

fun add_bounds_tac ctxt bound ts = let
    fun next (t, tac) = tac THEN' (add_bound_tac ctxt bound t)
in
    List.foldl next (K all_tac) ts 
end

fun trivial_insts thy cod_tab vars = 
    map(fn(x as Var(s,t))=>
	  (cterm_of thy x,
	   cterm_of thy (random_term' 1 0 0 thy cod_tab t))) vars

fun trig_thm thy thm = let
    val is_trig = Pattern.matches_subterm thy (@{term "trigger"}, prop_of thm)
                  orelse OldTerm.term_vars (prop_of thm) = []
in
    if is_trig then
        thm
    else
        thm RS @{thm "trigger_intr"} handle THM _ => thm
end

fun test_solve_tac ctxt n thm = let
    val te = TestEnv.get_data ctxt
    val cod_tab = #cod_term_tab(rep_testenv te)
in
    if Config.get ctxt SMT then let
            val thy = ProofContext.theory_of ctxt
	    val prems = prems_of thm
	    val prem = List.nth(prems,n-1)
	    val premvars = OldTerm.term_vars prem
            val datatype_vars = filter (fn x => isDataType thy (type_of x)) premvars
            val smt_facts = map (trig_thm thy) (TestEnv.get_smt_facts ctxt)
            val solve_tac = (add_bounds_tac ctxt (Config.get ctxt depth) datatype_vars)
                            THEN' (smt_ce_tac ctxt smt_facts)

            val insts = trivial_insts thy cod_tab datatype_vars 
	in
            Seq.single (instantiate ([],insts) (Seq.hd (solve_tac n thm)))
            handle SMT_Solver.SMT_COUNTEREXAMPLE (_,ce) 
                   => random_inst_tac ctxt cod_tab (SOME (mkinsts thy premvars ce)) n thm
		 | SMT_Solver.SMT _ => random_inst_tac ctxt cod_tab NONE n thm
	end
    else
	random_inst_tac ctxt cod_tab NONE n thm
end

fun solve_all ctxt cod_tab state = 
    state |$> (ALLCASES (test_solve_tac ctxt));


fun abs_data_solver ctxt atdata state = 
        state |$> ALLCASES(TRY'(SOLVE_ASMS ctxt (auto_solver atdata)))



  (* *********************************************************************** *)
  (* Normalizer                                                              *)
  (* *********************************************************************** *)
  (* motivation: type instantiation may make
     constraints of test-cases unsolvable. 
     Unsolvable cases should be removed before random-solving. 
     Since constraints with Mvars were not considered
     by simptac and arith_tac, they were substituted 
     against arbitrary free vars aforehand. *)

fun norm_tac ctxt n thm =
   let  val thy   = ProofContext.theory_of ctxt
        val prems = prems_of thm
        val m     = length prems
        val prem  = List.nth(prems,n-1)
	val k     = length (Logic.strip_imp_prems(prem)) 
        val goal  = trivial(cterm_of thy prem)
        val tvars = OldTerm.term_vars prem
        val insts = map(fn(x as Var((s,a),t))=>
                            (cterm_of thy x,
                             cterm_of thy (Free("XXX"^s^(Int.toString a),t))))
                       (tvars)
	val cleanup    = List.foldr (fn (_,tac) => (atac (n)) THEN tac)
			       all_tac (1 upto k)
   in  (
        (rtac (instantiate ([],insts) goal) n) 
        THEN 
        (full_simp_tac (simpset_of ctxt) n)  (* may erase goal n. 
		                 --> test case unsolvable. *) 
        THEN 
	(fn thm' => if (nprems_of thm' < m+k) (* goal erased *)
	            then cleanup thm'
		    else no_tac thm )(* report failure *)
       )(thm)
   end
    

fun norm_phase ctxt state = 
    state |$> (ALLCASES(TRY'(norm_tac ctxt)))


  (* *********************************************************************** *)
  (* core routine: gluing all local solvers together ...                     *)
  (* *********************************************************************** *)
  (* realizes the solver chain: first grounding,
     then erasing test cases that become unsolvable
     during grounding,
     then abstract test data resolution,
     then random solving. (could be enumeration instead).

     FUTURE: an arith-solver between abstract test data resolution and
     random solving. In particular sorting related tests would
     greatly profit from such a component.
   *)


fun gen_test_data_core ctxt name cdtab bound candTs testthm atdata =
    let  val thy = ProofContext.theory_of ctxt
         val solve_phase = solve_all ctxt cdtab
         val arith_phase  = fn thm => thm  (* not yet implemented *)
         val abs_data_phase= abs_data_solver ctxt atdata
         val type_grounds = groundT_thm thy bound candTs 
    in   map (solve_phase o arith_phase o abs_data_phase o (norm_phase ctxt))
             (type_grounds testthm)
    end


fun gen_test_data name context =
    let val ctxt = Context.proof_of context
        val te = TestEnv.get_data ctxt
        val bound = Config.get ctxt type_range_bound
        val candTs= #type_candicates(rep_testenv te)
        val testthm  = (the(Symtab.lookup(thm_tab te)name)
	             handle Option.Option => error("No test theorem: "^name^" available!"))
        val atdata= (the(Symtab.lookup(absdata_tab te)name)
                     handle Option.Option => [])
        val cdtab = #cod_term_tab(rep_testenv te)
        val iters = Config.get ctxt iterations
        val jdmts = gen_test_data_core ctxt name cdtab bound candTs testthm atdata
        val te' =
            jdmt_tab_update
               (Symtab.update(name,jdmts)(jdmt_tab te))
               te
    in TestEnv.map_data (K te') context end;

fun get_test_data (te,name) =
    maps (convert_goals_to_metahyps (not o is_thyp))
             (the(Symtab.lookup(jdmt_tab te)name)
	      handle Option.Option => error("No data statement: "^name^" available!"));


fun get_test_hyps (te,name) =
    maps (convert_goals_to_metahyps (is_thyp))
             (the(Symtab.lookup(jdmt_tab te)name)
	      handle Option.Option => error("No data statement: "^name^" available!"));


(* ************************************************************************* *)
(* micellaneous functions ...                                                *)
(* ************************************************************************* *)


fun animate thy (name, te) =  (* HACK - very simplistic implementation *)
    let fun animate_data_statement thm =
                (writeln ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>";
                 Pretty.writeln (Display.pretty_thm_global thy thm);
                 writeln "============================================";
   (*              asm_full_simplify (global_simpset_of(theory_of_thm(thm))) thm; *)
                 asm_full_simplify (global_simpset_of thy) thm; 
                 writeln "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
    in
       setmp_CRITICAL trace_simp true
         (fn () => List.app animate_data_statement (get_test_data(te,name))) ()
    end;

fun gen_testscript put smlput fname dataname thy =  let
    val te = TestEnv.get_data_global thy;

    val margin = 58;
    val put_term = Syntax.read_term_global thy put
    val put_is_const = is_Const put_term 
    val long_put = if put_is_const 
		   then let 
			   val Const(s,_) = put_term
		       in 
			   s
		       end
		   else put

    fun output_code gr xs = implode (map (snd o Graph.get_node gr)
					 (rev (Graph.all_preds gr xs)));
	

    fun output_code gr  xs =
	let
	    val module = ""
	    val code = List.mapPartial (fn s =>
					   let val c as (_, module', _) = Graph.get_node gr s
					   in if module = "" orelse module = module' then SOME (s, c) else NONE end)
				       (rev (Graph.all_preds gr xs));
	in
	    implode (map (#3 o snd) code)
	end;
	

    fun gen_generate_tc_code prep_term num thy  (speclist: (term list * term) list)  =  let 
	val numstr = Int.toString(num)
	fun image_type typ = (case typ of
				  Type(s,[t]) =>  Type(s,[t])
				| Type(@{type_name fun},(t::tl)) => image_type (hd (rev tl)) 
				| t => t);    
	val emptygr : Codegen.codegr = (Graph.empty,
				((Symtab.empty, Symtab.empty), (Symtab.empty, Symtab.empty)));

	val gen_wrapper = Config.get_thy thy wrapper;
	val gen_toString = Config.get_thy thy toString;
	val initGraph = Codegen.new_node ("<Top>", (NONE, "", "")) emptygr
	val gr = initGraph
	val smlput = if (smlput ="") then put else smlput
						   
	fun generate (speclist: (term list * term) list) = let
	    fun arity_of_type typ = (case typ of
					 Type(s,[]) =>  0
				       | Type(@{type_name fun},(t::tl)) => 1 + arity_of_type (hd tl)
				       | _ => 0
				    )
				    
	    fun arity_of src t = 
		case t of 
		    (Const(s,T)) => if (s=src) then (arity_of_type T) else 0
		  | (Free(s,T)) => if (s=src) then (arity_of_type T) else 0
		  | (Var(i,T)) => 0
		  | (Bound(i)) => 0
		  | (Abs(a,T,t)) => arity_of src t
		  | (f$t) => Int.max ((arity_of src f),(arity_of src t))
			     
	    val put_arity = if (gen_wrapper)
			    then List.foldl Int.max 0 (map (arity_of long_put) 
			                              (maps(fn(xs,y)=>y::xs) speclist))
			    else 0
	    fun wrap src t = 
		case t of 
		  (Const(@{const_name Char},T)$x$y) =>  (* Crude Hack for converting char literals *)
                      (Free("TestHarness.string_to_char",(Type (@{type_name char}, [])-->(Type (@{type_name char}, [])))))
						      $(Const(@{const_name Char},T)$(wrap src x)$(wrap src y)) 
		  | (Const(s,T)) 
		    => if (s=src) 
		       then if(gen_wrapper)
			    then Free("eval",T)
			    else Free(smlput,T) 
		       else Const(s,T) 
		  | (Free(s,T)) 
		    =>  if (s=src)  
			then if(gen_wrapper)
			     then Free("eval",T)
			     else Free(smlput,T)
			else Free(s,T)
		  | (Var(i,T)) => Var(i,T)
		  | (Bound(i)) => Bound(i)
		  | (Abs(a,T,t)) => Abs(a,T,wrap (src) t)
		  | (f$t) => (wrap (src) f )$ (wrap (src) t)
		  | t => t

	    fun gencode (gr, (t:term)) = swap (Codegen.invoke_codegen thy (Codegen.mk_deftab thy) "<Top>" ""
								      false (wrap long_put t) gr)

          fun to_code gr termlist = Library.foldl_map (gencode)  (gr, map (prep_term thy) termlist);
	  fun codeToString clist = (map (fn p => Pretty.string_of 
						     (Pretty.block [Pretty.brk 1, p])) clist);
	    fun code_spec (gr,(pre,post)) = let
		val (gr_pre, ps_pre) = to_code gr pre;
		val ps_pre = codeToString ps_pre
		val (gr_post,ps_post) = to_code gr_pre [post];
		val ps_post = codeToString ps_post
		    
	    in
		(gr_post, (ps_pre, hd ps_post))
	    end			    

	    val (gr', code) = Library.foldl_map (code_spec) (gr, speclist)
	
	    fun check_code (num,(pre,post)) = let 
		val numstr = Int.toString num;
		val num = num +1;
	    in
		(num, "val _ = print (\"\\nRunning Test Case "^numstr^":\\n\")\n"
                      ^"val pre_"^numstr
		      ^"  = [" ^
		      (space_implode ",\n             " 
				     (map (fn s => "fn () => (" ^ s ^ ")") pre)
		      ) 
		      ^ "];\n" ^
		      "val post_"^numstr^" = " ^
		      "fn () => (" ^ post ^ ")"
		      ^ ";\n"^
		      "val res_"^numstr^"  = TestHarness.check retval pre_"^numstr^
		      " post_"^numstr^";\n"^
		      "val testres = testres@[res_"^numstr^"];\n\n"
		)
	    end

	    val prelude_str = if Config.get_thy thy prelude
			      then   (output_code (fst gr')   ["<Top>"] )
			      else ""
	    val (num,check_str)   = Library.foldl_map (check_code) (num,code);
	    fun toString typ = (case typ of 
				    Type ("boolN", [])         => "Bool.toString"  (* FIXME "boolN" !?? *)
				  | Type (@{type_name int}, [])  => "Int.toString"
				  | Type(@{type_name real}, []) =>" Real.toString"
			 	  | Type(@{type_name list},[T]) => 
				    "(fn l => (\"[\"^(List.foldr (fn (s1,s2) \n"^
				    "                 => (if s2 = \"\" then s1 else s1^\", \"^s2))\n"^
				    "                    \"\" (map (fn x => "^(toString T)^
				    " x) l))^\"]\")) " 
				  | _ => "(fn s => \"\") ")
	    fun smltype_of typ = (case typ of 
				    Type ("boolN", [])       => SOME "bool"    (* FIXME "boolN" !?? *)
				  | Type (@{type_name int}, [])  => SOME "int"
				  | Type(@{type_name real}, []) => SOME "real"
			 	  | Type(@{type_name list},[T]) => (case smltype_of T of
				  	                           SOME t => SOME ("("^t^" list)")
							         | NONE => NONE)
				  | _ => NONE
			      ) 
	    val wrap_str = (if (gen_wrapper) 
			   then let
				   val cdtab = #cod_term_tab(rep_testenv te)   
				   val putT = if (put_is_const) 
					      then image_type (type_of (Syntax.read_term_global thy put))
					      else 
						  let 
						      fun isput (Free(s,t)) = if (s=put) 
									    then true 
									    else false
							| isput _ = false 
						      fun frees (a,b) = OldTerm.term_frees b
						      val postF = filter isput (maps frees speclist)
						      val Free(_,freetype) = hd postF
						  in 
						      image_type freetype
						  end
						  
				   val retval_term = random_term thy cdtab putT 
				   val (gr,ps) = to_code initGraph (* (Graph.new_node ("<Top>", (NONE, "")) 
									 Graph.empty) *) [retval_term];
				   val retval = hd (codeToString ps);
				   	      
						      
				   fun vars n =  if(n = 0) then "" 
						 else " x"^Int.toString(n)^" "^(vars (n-1))
				   val toStrings = if(gen_toString = "") 
						   then "fun toString  a = " ^(toString putT)^" a;"
						   else "";

				   val retvaltype = case (smltype_of putT) of
				   	              SOME t => ":("^t^")"
						    | NONE => ""
			       in
				   "val return = ref ("^retval^retvaltype^");\n"^
				   "fun eval "^(vars (put_arity))^" =  let val ret = "^smlput^" "
				   ^(vars (put_arity))^" in ((return := ret);ret)end\n"  ^
				   "fun retval () = SOME(!return);\n" ^
				   toStrings^"\n\n" 
			       end 
			   else
			       let
				   val toStrings = if(gen_toString = "") 
						   then  "fun toString a = \"\";"
						   else "";
			       in
				   "fun retval () = NONE;\n"^
				   toStrings^"\n\n"
 			       end)
			       
	 	val toStringName = if (gen_toString = "") then "toString" else gen_toString;
		    
	in  
	    
	    "(*******************************************************************\n"
	    ^" *                         Test-Driver \n"
			^" *           generated by HOL-TestGen "^testgen_version^"\n"
	    ^" *******************************************************************)\n\n"
	    ^"structure TestDriver : sig end = struct\n\n"
	    ^ Config.get_thy thy setup_code ^"\n\n"
	    ^prelude_str
	    ^ Config.get_thy thy dataconv_code ^"\n\n"
	    ^ wrap_str
	    ^ "val testres = [];\n\n"
	    ^ (List.foldl (fn (s1,s2) => s1^s2) "" check_str)
	    ^ "val _ = TestHarness.printList "^toStringName^" testres;\n\n"
	    ^ "end\n"
	end
    in		  
	PrintMode.setmp [] (Pretty.setmp_margin_CRITICAL margin generate) speclist
    end; 
	
	
    val generate_tc_code_i =  gen_generate_tc_code (K I); 
	
	
    val td = get_test_data (te,dataname);
    val prop_terms = map (fn p => let val _ $ ret = concl_of p  in ret end) td;  
    val prems_terms = map (fn data => (
			      map (fn p => let val _ $ ret = p in ret end) (prems_of data))) td;
	
    val spec = map2 (fn x => fn y => (x,y))  prems_terms prop_terms
    val tc   = generate_tc_code_i 0 thy spec;	   
    val _    = File.write (Path.explode fname) tc;
in
    thy
end;


(* module setup *)

val setup =
  trace_setup #>
  profiling_setup;

end (*struct *);

