
(*****************************************************************************
* HOL-TestGen --- theorem-prover based test case generation
*                 http://www.brucker.ch/projects/hol-testgen/
*                                                                            



* RBT_test_SMT_trigger.thy --- testing red-black trees with the Z3 solver
* Attention: This approach using triggers does not work very well.
* This file is part of HOL-TestGen.
*
* Copyright (c) 2005-2007, ETH Zurich, Switzerland
*
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are
* met:
*
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*
*     * Redistributions in binary form must reproduce the above
*       copyright notice, this list of conditions and the following
*       disclaimer in the documentation and/or other materials provided
*       with the distribution.
*
*     * Neither the name of the copyright holders nor the names of its
*       contributors may be used to endorse or promote products derived
*       from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
******************************************************************************)
(* $Id: RBT_test.thy 8967 2010-06-22 07:37:42Z wolff $ *)


header {* Testing Properties of Red-Black Trees *}


theory 
 RBT_test_SMT_trigger
imports 
 RBT_def
 Testing
begin

declare [[testgen_profiling]]
declare [[testgen_SMT]]

text{* The test plan is fairly standard and very similar to the AVL example: 
 test spec, test generation on the basis of some lemmas that allow for
 exploiting contradictions in constraints, data-generation and test script
 generation.

   Note that without the interactive proof part, the random solving phase
 is too blind to achieve a test script of suitable quality.
 Improving it will definitively improve also the quality 
 of the test. In this example, however, we deliberately stopped at the
 point where the quality was sufficient to produce relevant errors
 of the program under test.

   First, we define certain functions (inspired from the real implementation)
 that specialize the program to a sufficient degree: instead of generic trees
 over class @{term LINORDER}, we will generate test cases over integers.  

 *}

section {* Test Specification and Test-Case-Generation *}

instantiation int::LINORDER
begin

definition "compare (x::int) y
                 = (if (x < y) then LESS 
                                else (if (y < x) 
                                      then GREATER 
                                      else EQUAL))"

instance
  by intro_classes (simp_all add: compare_int_def)

end


lemma compare1[simp]:"(compare (x::int) y = EQUAL) = (x=y)"
by(auto simp:compare_int_def)

lemma compare2[simp]:"(compare (x::int) y = LESS) = (x<y)"
by(auto simp:compare_int_def)

lemma compare3[simp]:"(compare (x::int) y = GREATER) = (y<x)"
by(auto simp:compare_int_def)



text{* Now we come to the core part of the test generation:
  specifying the test specification. We will test an arbitrary program
  (insertion \verb+add+, deletion \verb+delete+) for test data
  that fulfills the following conditions:
  \begin{itemize}
  \item the trees must respect the invariants, i.e.~in particular
        the red and the black invariant,
  \item the trees must even respect the strong red invariant - i.e.
        the top node must be black,
  \item the program under test gets an additional parameter \verb+y+
        that is contained in the tree (useful for delete),
  \item the tree must be ordered (otherwise the implementations will
        fail). 
  \end{itemize}

  The analysis of previous test case generation attempts showed   
  that the following lemmas (altogether trivial to prove) help
  to rule out many constraints that are unsolvable - this knowledge
  is both useful for increasing the coverage (not so many failures
  will occur) as well for efficiency reasons: attempting to random solve
  unsolvable constraints takes time. Recall that that the number
  of random solve attempts is controlled by the \verb+iterations+
  variable in the test environment of this test specification. 
*}


lemma max_0_0 : "((max (a::nat) b) = 0) = (a = 0 \<and> (b = 0))"
 by(auto simp: max_def)

lemma [simp]: "(max (Suc a) b) ~= 0"
 by(auto simp: max_def)

lemma [simp]: "(max b (Suc a) ) ~= 0"
 by(auto simp: max_def)

lemma size_0[simp]: "(size x = 0) = (x = E)"
 by(induct "x",auto)

test_spec "(isord t \<and> isin (y::int) t \<and> 
            strong_redinv t \<and> blackinv t)
          \<longrightarrow> (blackinv(prog(y,t)))"

apply(gen_test_cases 5 1 "prog" simp: compare1 compare2 compare3)
store_test_thm "red-and-black-inv"

section {* Test Data Generation *}

lemma tree_exhaust: "trigger [pat (size x)] (x = E \<or> (\<exists> a b c d. x = T a b c d))"
apply(simp add: trigger_def)
apply(rule_tac y = "x" in  tree.exhaust)
apply(auto)
done

lemma tree_inject: "trigger [pat(T ?color ?tree1.0 ?a ?tree2.0) andpat (T ?color' ?tree1' ?a' ?tree2')] 
((T ?color ?tree1.0 ?a ?tree2.0 = T ?color' ?tree1' ?a' ?tree2') =
(?color = ?color' \<and> ?tree1.0 = ?tree1' \<and> ?a = ?a' \<and> ?tree2.0 = ?tree2'))"
apply(simp add: trigger_def)
done

lemma tree_distinct: "trigger [pat(T ?color' ?tree1' ?a' ?tree2')] (E \<noteq> T ?color' ?tree1' ?a' ?tree2')"
apply(simp add: trigger_def)
done

lemma tree_size_4: "size (T a b c d) = (size a) + (size b) + (size d) + Suc 0" 
apply(simp)
apply(rule_tac y = "a" in color.exhaust)
apply(auto)
done

lemma color_exhaust: "trigger [pat (size(x))] (x = B \<or> x = R)"
apply(simp add: trigger_def)
apply(rule_tac y = "x" in  color.exhaust)
apply(auto)
done

lemma isord_branch_alt1: "trigger [pat (isord (T ?c ?a ?y ?b))] (isord (T ?c ?a ?y ?b) \<longrightarrow> (isord ?a \<and>
 isord ?b))"
apply(simp add: trigger_def)
apply(auto)
done

lemma isord_branch_alt2:  "trigger [pat (isord (T ?c ?a ?y ?b)) andpat (T u t1 x t2)] 
(isord (T ?c ?a ?y ?b) \<longrightarrow> (isin x ?a \<longrightarrow> compare x ?y = LESS))"
apply(simp add: trigger_def)
apply(auto)
done

lemma isord_branch_alt3:  "trigger [pat (isord (T c a y b)) andpat (T u t1 x t2)] 
(isord (T c a y b) \<longrightarrow> (isin x b \<longrightarrow> compare x y = GREATER))"
apply(simp add: trigger_def)
done

lemma isord_branch_alt: "trigger [pat (isord (T ?c ?a ?y ?b))] (isord (T ?c ?a ?y ?b) =
(isord ?a \<and>
 isord ?b \<and>
 trigger [pat(T c t1 x t2)] (\<forall>x. isin x ?a \<longrightarrow> compare x ?y = LESS) \<and>
 trigger [pat(T c t1 x t2)] (\<forall>x. isin x ?b \<longrightarrow> compare x ?y = GREATER)))"
by(simp add: trigger_def isord_branch)

lemmas max_def_2 = "Orderings.ord_class.max_def"[where 'a = nat]

declare max_def_2 [testgen_smt_facts]

declare tree_inject [testgen_smt_facts]
declare tree_distinct [testgen_smt_facts]
declare tree.size(3) [testgen_smt_facts]
declare tree_size_4 [testgen_smt_facts]
declare tree_exhaust [testgen_smt_facts]

declare color.distinct(1) [testgen_smt_facts]
declare color.size(3) [testgen_smt_facts]
declare color.size(4) [testgen_smt_facts]
declare color_exhaust [testgen_smt_facts]

(* declare isord.simps [testgen_smt_facts] *)
declare isord_empty [testgen_smt_facts]
declare isord_branch_alt1 [testgen_smt_facts]
declare isord_branch_alt2 [testgen_smt_facts]
declare isord_branch_alt3 [testgen_smt_facts]
declare isin.simps [testgen_smt_facts]
declare RBT_test_SMT_trigger.compare_int_def [testgen_smt_facts]
declare blackinv.simps [testgen_smt_facts]
declare max_B_height.simps [testgen_smt_facts]
declare strong_redinv.simps [testgen_smt_facts]
declare redinv.simps [testgen_smt_facts]

thm testgen_smt_facts

declare [[ z3_options = "AUTO_CONFIG=false" ]]

gen_test_data "red-and-black-inv"

thm "red-and-black-inv.test_data"

end